import sys
from PyQt5.QtWidgets import QTabWidget, QLabel, QHeaderView, QFrame, QTableWidget, QTableWidgetItem, QWidget, QGraphicsWidget, QMessageBox, QErrorMessage, QInputDialog, QLineEdit, QApplication, QGraphicsScene, QGraphicsView, QGraphicsProxyWidget, QTextEdit, QVBoxLayout, QHBoxLayout, QGraphicsRectItem, QGraphicsLineItem, QDialog, QSizePolicy, QListWidget, QPushButton
from PyQt5.QtCore import QPointF, Qt, QLineF, QSize, QRectF
from PyQt5.QtGui import QResizeEvent, QFontMetrics


from claid import CLAID
from claid.dispatch.proto.claidservice_pb2  import CLAIDConfig, ModuleAnnotation, DataPackage


class LogMessageView(QWidget):

    def __init__(self):
        super().__init__()




from claid.dispatch.proto.claidservice_pb2  import CLAIDConfig, ModuleAnnotation, DataPackage
from google.protobuf.json_format import Parse, ParseDict, MessageToJson
import json
from config.channel_description import ChannelDescription

class Config():

    def __init__(self):
        self.config = None

    def load_config_from_file(self, json_file_path):
        self.config = CLAIDConfig()
        with open(json_file_path, 'r') as f:
            json_data = json.load(f)
            # Use ParseDict to populate the protobuf message from a dictionary
            ParseDict(json_data, self.config)
        return self.config

    def save_config_to_file(self, path):
        Config.save_config_to_file(self.config, path)

    #TODO: Static methods, etc. recherchieren
    @staticmethod
    def save_config_to_file(config, json_file_path):
        # Use the .to_dict() method to convert the protobuf message to a dictionary
        json_data = MessageToJson(config, including_default_value_fields=True, preserving_proto_field_name=True)
        with open(json_file_path, 'w') as f:
            f.write(json_data)

    def get_config_data(self):
        return self.config
    
    def get_channel_descriptions(self):
        descriptions = dict()

        for host in self.config.hosts:
            for module in host.modules:
                input_channels = module.input_channels
                output_channels = module.output_channels

                for input_channel_name in input_channels:
                    input_channel_connection = input_channels[input_channel_name]

                    if not input_channel_connection in descriptions:
                        descriptions[input_channel_connection] = ChannelDescription(input_channel_connection, list(), list())

                    descriptions[input_channel_connection].subscriber_modules.append((host.hostname, module.id, input_channel_name))

                for output_channel_name in output_channels:
                    output_channel_connection = output_channels[output_channel_name]

                    if not output_channel_connection in descriptions:
                        descriptions[output_channel_connection] = ChannelDescription(output_channel_connection, list(), list())

                    descriptions[output_channel_connection].publisher_modules.append((host.hostname, module.id, output_channel_name))

        return descriptions






import sys
from PyQt5.QtWidgets import QTabWidget, QLabel, QHeaderView, QFrame, QTableWidget, QTableWidgetItem, QWidget, QGraphicsWidget, QMessageBox, QErrorMessage, QInputDialog, QLineEdit, QApplication, QGraphicsScene, QGraphicsView, QGraphicsProxyWidget, QTextEdit, QVBoxLayout, QHBoxLayout, QGraphicsRectItem, QGraphicsLineItem, QDialog, QSizePolicy, QListWidget, QPushButton
from PyQt5.QtCore import QPointF, Qt, QLineF, QSize, QRectF, pyqtSignal
from PyQt5.QtGui import QResizeEvent, QFontMetrics
from ui.module_graphics_item import ModuleGraphicsItem

from ui.non_scrollable_graphics_view import NonScrollableGraphicsView

from claid import CLAID
from claid.dispatch.proto.claidservice_pb2  import CLAIDConfig, ModuleAnnotation, DataPackage

from ui.property_input_dialog import PropertyInputDialog

class ModuleCatalog(QWidget):

    on_catalog_closed = pyqtSignal()

    def __init__(self, parent_host, claid, host_name, module_injections_supported=False):

        super().__init__()

        self.parent_host = parent_host
        self.module_injections_supported = module_injections_supported
        self.claid = claid


        self.available_modules = self.claid.get_available_modules_for_host(host_name)
        if self.available_modules == None:
            self.close_catalog()
        self.available_modules = dict(sorted(self.available_modules.items()))
        # Set up the main layout
        main_layout = QVBoxLayout(self)

        # Create a QGraphicsView and QGraphicsScene
        self.graphics_view = NonScrollableGraphicsView(self)
        self.scene = QGraphicsScene(self)
        self.graphics_view.setScene(self.scene)

        # Dimensions
        initial_width = 600
        initial_height = 600
        
        
        self.ratio_module_text = 0.3

        self.module_width = initial_width/4
        self.module_height = self.module_width * 2/3

        # Create a rectangle above the text
        self.rect_item = ModuleGraphicsItem("Test", "Test", 0, 0, self.module_width, self.module_height, [], [], [], [], [], [], [], movable=False)
        #! Should we delete the setPos code of as it is set in ResizeEvent() anyways?
        self.rect_item.setPos(initial_width//2 - self.module_width//2, (initial_height * self.ratio_module_text)//2 - self.module_height//2)
        self.scene.addItem(self.rect_item)

        # Create a horizontal line to separate text and rectangle
        #! Should we delete the arguments of QGraphicsLineItem as they are set in ResizeEvent() anyways?
        self.line_item = QGraphicsLineItem(0, initial_height * self.ratio_module_text, initial_width, initial_height * self.ratio_module_text)
        self.scene.addItem(self.line_item)

        # Create a QTextEdit with non-editable text
        self.module_text = QTextEdit()
        self.module_text.setPlainText("Empty")
        self.module_text.setReadOnly(True)
        self.module_text.setFrameShape(QTextEdit.NoFrame)
        self.module_text.setStyleSheet("border: none;")
        self.module_text.setFixedHeight(self.calculate_text_height("Empty"))
        # self.module_text.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        # Create a QGraphicsProxyWidget to embed the QTextEdit in the QGraphicsScene

        proxy_widget_inner = QWidget()
        proxy_widget_inner.setLayout(QVBoxLayout())
        proxy_widget_inner.layout().addWidget(self.module_text)
        self.proxy_widget = self.scene.addWidget(proxy_widget_inner)
        proxy_widget_inner.setStyleSheet("border: none;")

        proxy_widget_inner.layout().setSpacing(20)
        proxy_widget_inner.layout().setContentsMargins (0, 0, 0, 0)
        proxy_widget_inner.setStyleSheet("border: 0px")

        proxy_widget_inner.layout().addWidget(QLabel("Properties"))


        self.properties_table = QTableWidget()
        # Add necessary initialization for the first table
        self.properties_table.setRowCount(2)
        self.properties_table.setColumnCount(2)
        for i in range(2):
            for j in range(2):
                item = QTableWidgetItem(f"properties_table Item {i}, {j}")
                self.properties_table.setItem(i, j, item)

        # Hide grid lines
        self.properties_table.horizontalHeader().setSectionResizeMode(0, QHeaderView.ResizeToContents)
        self.properties_table.horizontalHeader().setSectionResizeMode(1, QHeaderView.Stretch)
        # Remove frame
        self.properties_table.setFrameStyle(QFrame.NoFrame)
        self.properties_table.setWordWrap(True)
        proxy_widget_inner.layout().addWidget(self.properties_table)
        self.properties_table.horizontalHeader().sectionResized.connect(self.properties_table.resizeRowsToContents)

        proxy_widget_inner.layout().addWidget(QLabel("Channels"))
        self.channels_table = QTableWidget()
        # Add necessary initialization for the first table
        self.channels_table.setRowCount(2)
        self.channels_table.setColumnCount(2)
        for i in range(2):
            for j in range(2):
                item = QTableWidgetItem(f"channels_table Item {i}, {j}")
                self.channels_table.setItem(i, j, item)

        # Hide grid lines
        
        # Remove frame
        self.channels_table.setFrameStyle(QFrame.NoFrame)
        self.channels_table.setWordWrap(True)
        self.channels_table.horizontalHeader().sectionResized.connect(self.channels_table.resizeRowsToContents)
        proxy_widget_inner.layout().addWidget(self.channels_table)

        # Add the QGraphicsView to the main layout
        main_layout.addWidget(self.graphics_view)

        # Set the main layout for the QDialog
        self.setLayout(main_layout)
        QApplication.instance().installEventFilter(self)

        # Remove scroll bars
        self.graphics_view.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        #self.graphics_view.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)


        modules_by_categories = dict()
        self.module_categories = list()

        for module_name in self.available_modules:

            module_annotation = self.available_modules[module_name]

            category = module_annotation.module_category
            if category == "":
                category = "Standard"

            
            if category not in modules_by_categories:
                modules_by_categories[category] = list()

            modules_by_categories[category].append(module_name)
            self.module_categories.append(category)

        self.tabwidget = QTabWidget()
        self.tabwidget.setMaximumHeight(200)  #  the value as needed
        self.list_widgets_per_category = dict()
        for category_index, module_category in enumerate(modules_by_categories):
            module_list = [module_name for module_name in modules_by_categories[module_category]]
            # Create a list widget

            list_widget = QListWidget(self)
            list_widget.addItems(module_list)  # Add some example items
            list_widget.setMaximumHeight(200)  #  the value as needed
            # Connect item selection changed signal to a function
            list_widget.itemSelectionChanged.connect(self.handle_list_selection)
            list_widget.setCurrentRow(0)


            self.tabwidget.addTab(list_widget,  module_category)
            self.list_widgets_per_category[category_index] = list_widget

        self.tabwidget.setCurrentIndex(0)
        # Create layout for the QGraphicsView
        

        # Create Ok and Cancel buttons
        ok_button = QPushButton("Ok", self)
        cancel_button = QPushButton("Cancel", self)

        main_layout.addWidget(QLabel("Available Modules"))
        main_layout.addWidget(self.tabwidget)
        buttons_layout = QHBoxLayout()
        buttons_layout.addWidget(ok_button)
        buttons_layout.addWidget(cancel_button)

        ok_button.clicked.connect(self.on_ok_button_clicked)
        cancel_button.clicked.connect(self.close_catalog)

        main_layout.addLayout(buttons_layout)

        # Set layout for the dialog
        self.setLayout(main_layout)
        self.update()
        self.graphics_view.setInteractive(False)

        self.resize(initial_width, initial_height)
        self.proxy_widget.widget().resize(self.graphics_view.width(), self.graphics_view.height())

        # self.on_module_selected(module_list[0])
        self.update()



    def scrollContentsBy(self, dx, dy):
        # Prevent horizontal scrolling
        print("Scroll")
        super().scrollContentsBy(0, dy)

    def resizeEvent(self, event):
        print("resize event")
        # Adjust the line's length to match the width of the QTextEdit
        print(event)

        new_size = event.size()

        # Extract width and height values
        width = new_size.width()
        height = new_size.height()

        min_size = 500
        print(width, height)
        print(width, height)
        if(width < min_size):
            width = min_size
            print("resize")
            self.resize(min_size, height)
            return
        
        if(height < min_size):
            height = min_size
            print("resize")
            self.resize(width, min_size)
            return

        y_pos = height * self.ratio_module_text


        print(self.graphics_view.height())
        self.line_item.setLine(0, y_pos, self.graphics_view.width(), y_pos)

        self.proxy_widget.setPos(0, y_pos + 10)  # Adjust the position as needed
        # self.proxy_widget.setRect(self.graphics_view.width(), self.graphics_view.height() * self.ratio_module_text)
        # self.proxy_widget.setGeometry(QRectF(0,height * self.ratio_module_text + 10, self.graphics_view.width(), height))

        self.module_text.setFixedHeight(self.calculate_text_height(self.module_text.toPlainText()))

        self.proxy_widget.widget().resize(self.graphics_view.width(), self.proxy_widget.widget().height())

        self.rect_item.setPos(self.graphics_view.width()//2 - self.rect_item.rect().width()//2, y_pos // 2 - self.module_height//2)

    def handle_list_selection(self):
        idx = self.tabwidget.currentIndex()
        if(idx == -1):
            return
        selected_item = self.list_widgets_per_category[idx].currentItem()
        if selected_item:
            # Call a function with the selected item's text
            self.on_module_selected(selected_item.text())

        print("list selection done")

    def on_module_selected(self, item_text):
        # You can fill out this function with your custom logic
        print(f"Selected item: {item_text}")
        self.show_information_for_module(item_text)

    def show_information_for_module(self, module_type: str):
        print("show info")
        module_annotation = self.available_modules[module_type]
        input_channels = self.claid.get_input_channels_of_module(module_annotation)
        output_channels = self.claid.get_output_channels_of_module(module_annotation)

        input_channel_types = self.claid.get_input_channel_types_of_module(module_annotation)
        output_channel_types = self.claid.get_output_channel_types_of_module(module_annotation)

        description = module_annotation.module_description
        # description = "Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet. Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet. Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet.   uis autem vel eum iriure dolor in hendrerit in vulputate velit esse molestie consequat, vel illum dolore eu feugiat nulla facilisis at vero eros et accumsan et iusto odio dignissim qui blandit praesent luptatum zzril delenit augue duis dolore te feugait nulla facilisi. Lorem ipsum dolor sit amet,"
        self.module_text.setText(description)
        self.module_text.setFixedHeight(self.calculate_text_height(description))

        self.generate_property_descriptions(module_annotation)
        self.generate_channel_descriptions(module_annotation)
        

        self.rect_item.set_channels(input_channels, output_channels, input_channel_types, output_channel_types)
        self.rect_item.set_type(module_type)
        self.rect_item.set_id("New Module")
        print("show info done")
        
    def generate_property_descriptions(self, module_annotation):
        
        self.properties_table.setRowCount(len(module_annotation.properties))
        self.properties_table.setColumnCount(2)

        self.properties_table.setHorizontalHeaderLabels(["Property", "Description"])

        ctr = 0

        for row in range(len(module_annotation.properties)):
            for col in range(2):
                if self.properties_table.item(row, col):
                    self.properties_table.item(row, col).setText("")
                else:
                    new_item = QTableWidgetItem("")
                    self.properties_table.setItem(row, col, new_item)

        for property, description in zip(module_annotation.properties, module_annotation.property_descriptions):
            self.properties_table.item(ctr, 0).setText(property)
            self.properties_table.item(ctr, 1).setText(description)
            self.properties_table.update()
            ctr += 1
            
    def generate_channel_descriptions(self, module_annotation):

        channels = module_annotation.channel_definition

        self.channels_table.setRowCount(len(channels))
        self.channels_table.setColumnCount(4) # Channel name, channel type, channel data type, channel description

        ctr = 0

        input_channels = list()
        input_channel_descriptions = list()
        output_channels = list()
        output_channel_descriptions = list()

        data_types_dict = dict()

        for channel, description in zip(module_annotation.channel_definition, module_annotation.channel_description):
            
            print(channel)
            data_types_dict[channel.channel] = channel.WhichOneof("payload_oneof")
            if channel.source_module != "":
                # Is output channel
                output_channels.append(channel.channel)
                output_channel_descriptions.append(description)

            elif channel.target_module != "":
                # Is input channel
                input_channels.append(channel.channel)
                input_channel_descriptions.append(description)

            else:
                raise ValueError("Invalud channel {}, cannot determine whether it is an input or output channel since neither source nor target are set.".format(channel.channel))

        for row in range(len(channels)):
            for col in range(4):
                if self.channels_table.item(row, col):
                    self.channels_table.item(row, col).setText("")
                else:
                    new_item = QTableWidgetItem("")
                    self.channels_table.setItem(row, col, new_item)
        

        for channel, description in zip(input_channels, input_channel_descriptions):
            self.channels_table.item(ctr, 0).setText(channel)
            self.channels_table.item(ctr, 1).setText("Input")
            self.channels_table.item(ctr, 2).setText(data_types_dict[channel])
            self.channels_table.item(ctr, 3).setText(description)
            print("ctr", ctr)
            ctr += 1

        for channel, description in zip(output_channels, output_channel_descriptions):
            self.channels_table.item(ctr, 0).setText(channel)
            self.channels_table.item(ctr, 1).setText("Output")
            self.channels_table.item(ctr, 2).setText(data_types_dict[channel])
            self.channels_table.item(ctr, 3).setText(description)
            print("ctr", ctr)
            ctr += 1

        self.channels_table.setHorizontalHeaderLabels(["Channel", "Channel type", "Data type", "Description"])

        self.channels_table.horizontalHeader().setSectionResizeMode(0, QHeaderView.ResizeToContents)
        self.channels_table.horizontalHeader().setSectionResizeMode(1, QHeaderView.ResizeToContents)
        self.channels_table.horizontalHeader().setSectionResizeMode(2, QHeaderView.ResizeToContents)
        self.channels_table.horizontalHeader().setSectionResizeMode(3, QHeaderView.Stretch)
        self.channels_table.update()

    def on_ok_button_clicked(self):

        module_name, ok = QInputDialog().getText(self, "Module name",
                                     "Enter a unique name for the Module.", QLineEdit.Normal)
        if not ok or not module_name:
            return
        
        if self.parent_host.does_module_with_name_exist(module_name):
            QMessageBox.critical(None, "Error", "A Module with name \"{}\" already exists. Please choose an unique name for the Module.".format(module_name))

            return 

        idx = self.tabwidget.currentIndex()
        if(idx == -1):
            return
        
        list_widget = self.list_widgets_per_category[idx]
        module_type = list_widget.currentItem().text()
        module_annotation = self.available_modules[module_type]
        input_channels = self.claid.get_input_channels_of_module(module_annotation)
        output_channels = self.claid.get_output_channels_of_module(module_annotation)
        input_channel_types = self.claid.get_input_channel_types_of_module(module_annotation)
        output_channel_types = self.claid.get_output_channel_types_of_module(module_annotation)

        properties = module_annotation.properties
        property_descriptions = module_annotation.property_descriptions
        property_hints = module_annotation.property_hints

        empty_properties = [42]

        self.defined_properties = dict()

        self.__property_dialog_module_name = module_name
        self.property_dialog = PropertyInputDialog(properties, property_descriptions, property_hints)
        self.property_dialog.onOkButtonClicked.connect(self.onPropertyDialogOkClicked)

        if len(properties) > 0:
            self.property_dialog.open()
        else:
            self.onPropertyDialogOkClicked()

    def onPropertyDialogOkClicked(self):

        idx = self.tabwidget.currentIndex()
        if(idx == -1):
            return
        
        list_widget = self.list_widgets_per_category[idx]
        module_type = list_widget.currentItem().text()
        module_annotation = self.available_modules[module_type]
        input_channels = self.claid.get_input_channels_of_module(module_annotation)
        output_channels = self.claid.get_output_channels_of_module(module_annotation)
        input_channel_types = self.claid.get_input_channel_types_of_module(module_annotation)
        output_channel_types = self.claid.get_output_channel_types_of_module(module_annotation)

        properties = module_annotation.properties
        property_descriptions = module_annotation.property_descriptions
        property_hints = module_annotation.property_hints

        self.defined_properties = self.property_dialog.get_defined_properties()
        empty_properties = self.property_dialog.get_empty_properties()

        if len(empty_properties) != 0:

            empty_list_msg = ""
            for property in empty_properties:
                empty_list_msg += "\t\"" + str(property) +  "\"\n"
            QMessageBox.critical(None, "Undefined property", "You did not specify a value for the following properties:\n{}".format(empty_list_msg) + 
                                    "Please make sure that you define each property.")

            self.property_dialog = PropertyInputDialog(properties, property_descriptions, property_hints, self.defined_properties)
            self.property_dialog.onOkButtonClicked.connect(self.onPropertyDialogOkClicked)
            self.property_dialog.open()
            return

        module = self.parent_host.add_module(self.__property_dialog_module_name, module_type, 0, 0, input_channels, output_channels,\
                                              input_channel_types, output_channel_types, properties, property_descriptions, property_hints)
        module.set_properties(self.defined_properties)
        
        if self.module_injections_supported:
            module.set_is_injectable(module_annotation.is_injectable)
            module.set_injection_file_dependencies(module_annotation.file_dependencies)
        self.close_catalog()
    
    def close_catalog(self):
        self.close()
        self.on_catalog_closed.emit()


    def calculate_text_height(self, text):
        document = self.module_text.document()
        docHeight = document.size().height()
        print("HEIGHT ",
               docHeight)
        return int(docHeight)
    
 


from PyQt5.QtWidgets import QGraphicsObject, QGraphicsSceneHoverEvent, QGraphicsSceneMouseEvent
from PyQt5.QtCore import QRectF, pyqtSignal, Qt, QEvent
from PyQt5.QtGui import QPen, QColor, QPainterPath


class Channel(QGraphicsObject):

    onChannelClicked = pyqtSignal(QGraphicsObject)

    def __init__(self, name: str, data_type, x: int, y: int, diameter: int, parent=None):
        #TODO: Cannot import ModuleGraphicsItem due to circular import issue
        # if parent == None or not isinstance(parent, ModuleGraphicsItem):
        #     raise ValueError("Parent of Channel has to be a ModuleGraphicsItem, and cannot be None")
        if parent != None:
            super(Channel, self).__init__(parent)
        else:
            super(Channel, self).__init__()
    
        self.module_parent = parent
        # Initialize major channel attributes
        self.name = name
        self.data_type = data_type

        # self.data_type = data_type
        print("Channel 2")

        # Set dimensions
        self.diameter = diameter

        # Set colors and thickness
        self.thickness = 2
        self.default_color = self.parentItem().get_color()
        self.default_pen = QPen(self.default_color, self.thickness)
        self.default_brush = QColor(self.default_color)
        self.hover_color = QColor("#548CCC")
        self.hover_pen = QPen(self.hover_color, self.thickness * 1.5)
        self.hover_brush = QColor(self.hover_color)
        self.click_color = QColor("#1A5CA8")
        self.click_pen = QPen(self.click_color, self.thickness * 1.5)
        self.click_brush = QColor(self.click_color)
        print("Channel 3")

        # Set position:
        self.setPos(x, y)

        # Check if it is input channel
        if x == 0:
            self.is_input = True
        else:
            self.is_input = False

        # Draw channel shape at initialization
        self.path = self.drawChannel()

        # Set flags
        self.setAcceptHoverEvents(True)
        self.hovered = False
        self.clicked = False
        print("Channel 4")


    def boundingRect(self):
        return self.path.boundingRect()

    def paint(self, painter, option, widget=None):
        # Set color and brush
        if self.isClicked():
            channel_pen = self.click_pen
            channel_brush = self.click_brush
        elif self.isHovered():
            channel_pen = self.hover_pen
            channel_brush = self.hover_brush
        else:
            channel_pen = self.default_pen
            channel_brush = self.default_brush
        painter.setPen(channel_pen)
        painter.setBrush(channel_brush)

        # Draw channel
        painter.drawPath(self.path)

    def hoverEnterEvent(self, event):
        self.hovered = True
        if self.parentItem().isHovered():
            self.parentItem().set_hovered(False)
        self.update()
        super(Channel, self).hoverEnterEvent(event)

    def hoverLeaveEvent(self, event):
        self.hovered = False
        self.update()
        super(Channel, self).hoverLeaveEvent(event)

    def mousePressEvent(self, event):
        # Clear selection of other GraphicsItems
        if not self.isSelected():
            self.scene().clearSelection()

        if event.button() == Qt.LeftButton:
            self.clicked = True
            self.update()
            self.onChannelClicked.emit(self)
        else:
            super(Channel, self).mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.clicked = False
            self.update()
        super(Channel, self).mouseReleaseEvent(event)

    def drawChannel(self) -> QPainterPath:
        path = QPainterPath()
        path.moveTo(0, 0)
        if self.is_input:
            path.lineTo((-self.diameter * 2/12), 0)
            path.arcTo((-self.diameter * 5/12), 0, (self.diameter * 6/12), self.diameter, 90, 180)
        else:
            path.lineTo((self.diameter * 2/12), 0)
            path.arcTo((-self.diameter * 1/12), 0, (self.diameter * 6/12), self.diameter, 90, -180)
        path.lineTo(0, self.diameter)
        path.closeSubpath()

        return path

    def get_name(self) -> str:
        return self.name

    def has_same_data_type_as_channel(self, other_channel):
        

        our_data_type_case = self.data_type.WhichOneof("payload_oneof")
        other_data_type_case = other_channel.get_data_type().WhichOneof("payload_oneof")

        print(our_data_type_case)
        print(other_data_type_case)

        # Temporary exception for the DataSaver
        if self.module_parent.get_type() == "DataSaverModule" or other_channel.module_parent.get_type() == "DataSaverModule":
            return True

        if our_data_type_case != other_data_type_case:
            return False
        
        if our_data_type_case == "blob_val":
            our_message_type = self.data_type.blob_val.message_type
            other_message_type = other_channel.get_data_type().blob_val.message_type

            if our_message_type != other_message_type:
                return False
            
        
        return True
    

    def get_data_type(self):
        return self.data_type
    
    def get_data_type_name(self):
        name = self.data_type.WhichOneof("payload_oneof")

        if name != "blob_val":
            return name
        
        return self.data_type.blob_val.message_type
    
    def get_name(self):
        return self.name

    def isClicked(self) -> bool:
        return self.clicked
    
    def isHovered(self) -> bool:
        return self.hovered
    
    def set_hovered(self, hovered: bool) -> None:
        self.hovered = hovered
        self.update()

    def set_default_color(self, color: QColor) -> None:
        self.default_color = color
        self.default_pen = QPen(self.default_color, self.thickness)
        self.default_brush = QColor(self.default_color)
        # self.update()



import sys
from PyQt5.QtWidgets import QFileDialog, QComboBox, QLabel, QHeaderView, QFrame, QTableWidget, QTableWidgetItem, QWidget, QGraphicsWidget, QMessageBox, QErrorMessage, QInputDialog, QLineEdit, QApplication, QGraphicsScene, QGraphicsView, QGraphicsProxyWidget, QTextEdit, QVBoxLayout, QHBoxLayout, QGraphicsRectItem, QGraphicsLineItem, QDialog, QSizePolicy, QListWidget, QPushButton
from PyQt5.QtCore import QPointF, Qt, QLineF, QSize, QRectF, pyqtSignal
from PyQt5.QtGui import QResizeEvent, QFontMetrics, QIntValidator

from ui.non_scrollable_graphics_view import NonScrollableGraphicsView

from claid.dispatch.proto.claidservice_pb2  import PropertyHint, PropertyType
import os
class PropertyInputDialog(QDialog):

    onOkButtonClicked = pyqtSignal()

    def __init__(self, properties, property_descriptions, property_hints, defined_properties = None):

        super().__init__()

        self.properties = properties
        self.property_descriptions = property_descriptions
        self.property_hints = property_hints

        if defined_properties == None:
            defined_properties = dict()

        # Set up the main layout
        main_layout = QVBoxLayout(self)

        self.__property_input_ui_elements = dict()
        self.__property_path_buttons = dict()

        self.was_cancelled = False

        main_layout.addWidget(QLabel("Please configure the properties of the Module.\nProperties:"))

        for property, description, property_hint in zip(properties, property_descriptions, property_hints):
            layout = QHBoxLayout()

            label = QLabel(property)
            label.setToolTip(description)
            layout.addWidget(label)

            property_ui_element = self.__add_property_ui_elements_according_to_property_type(property, description, property_hint, defined_properties)
            
            if isinstance(property_ui_element, list):
                for element in property_ui_element:
                    layout.addWidget(element)
            else:
                layout.addWidget(property_ui_element)
            main_layout.addLayout(layout)

            self.__property_input_ui_elements[property] = property_ui_element


        # Create Ok and Cancel buttons
        ok_button = QPushButton("Ok", self)
        cancel_button = QPushButton("Cancel", self)
        ok_button.clicked.connect(self.on_ok_button_clicked)
        cancel_button.clicked.connect(self.on_cancel_button_clicked)

        buttons_layout = QHBoxLayout()
        buttons_layout.addWidget(ok_button)
        buttons_layout.addWidget(cancel_button)
        main_layout.addLayout(buttons_layout)

        # Set layout for the dialog
        self.setLayout(main_layout)

    def __add_property_ui_elements_according_to_property_type(self, property: str, property_description: str, property_hint: PropertyHint, defined_properties: dict):
        
        if property_hint.property_type == PropertyType.PROPERTY_TYPE_DEFAULT:
        
            line_edit = QLineEdit()
            line_edit.setToolTip(property_description)
            
            if property in defined_properties:
                line_edit.setText(defined_properties[property])

            return line_edit
        
        elif property_hint.property_type == PropertyType.PROPERTY_TYPE_ENUM:
            combobox = QComboBox()
            combobox.addItems(property_hint.property_type_enum_values)

            if property in defined_properties:
                combobox.setCurrentText(defined_properties[property])

            return combobox
        
        elif property_hint.property_type == PropertyType.PROPERTY_TYPE_INT:
        
            line_edit = QLineEdit()
            line_edit.setToolTip(property_description)
            validator = QIntValidator(property_hint.property_type_int_min, property_hint.property_type_int_max)
            line_edit.setValidator(validator)
            if property in defined_properties:
                line_edit.setText(defined_properties[property])

            return line_edit
        
        elif property_hint.property_type == PropertyType.PROPERTY_TYPE_PATH:
        
            horizontal_layout = QHBoxLayout()
            line_edit = QLineEdit()
            line_edit.setToolTip(property_description)
            
            if property in defined_properties:
                line_edit.setText(defined_properties[property])


            button = QPushButton("...")
            button.clicked.connect(lambda _, button=button: self.on_path_button_clicked(button))
            self.__property_path_buttons[button] = property
            return [line_edit, button]
        

    def get_defined_properties(self):

        properties = dict()
        for property in self.properties:

         
            value = self.__get_property_value_from_ui_element(property)

            if value != "":
                properties[property] = value

        return properties
    
    def get_empty_properties(self):

        empty_properties = list()
        for property in self.properties:
            
            value = self.__get_property_value_from_ui_element(property)

            if value == "":
                empty_properties.append(property)

        return empty_properties

    def __get_property_value_from_ui_element(self, property):

        if property not in self.__property_input_ui_elements:
            return ""
        
        ui_element = self.__property_input_ui_elements[property]

        property_hint = None

        for i in range(len(self.properties)):
            if self.properties[i] == property:
                property_hint = self.property_hints[i]

        if property_hint == None:
            raise ValueError("Failed to get property_hint of property \"{}\"".format(property))
        property_type = property_hint.property_type

        if property_type == PropertyType.PROPERTY_TYPE_DEFAULT or property_type == PropertyType.PROPERTY_TYPE_INT:
            return ui_element.text()

        elif property_type == PropertyType.PROPERTY_TYPE_ENUM:
            return ui_element.currentText()
        
        elif property_type == PropertyType.PROPERTY_TYPE_PATH:
            return ui_element[0].text()

    def on_ok_button_clicked(self):

        if not self.verify_properties():
            return
        self.was_cancelled = False
        self.close()
        self.onOkButtonClicked.emit()

    def verify_properties(self):
        
        for property, property_hint in zip(self.properties, self.property_hints):
            if property_hint.property_type == PropertyType.PROPERTY_TYPE_INT:
                line_edit = self.__property_input_ui_elements[property]

                input_text = line_edit.text()
                try:
                    number = int(input_text)

                    if number < property_hint.property_type_int_min or number > property_hint.property_type_int_max:
                        QMessageBox.critical(None, "Out of range", "Value {} of property \"{}\" is out of range [{},{}]. Please choose a value within the range.".format(\
                            number, property, property_hint.property_type_int_min, property_hint.property_type_int_max))
                        return False
                except Exception as e:
                    QMessageBox.critical(None, "Out of range", "Value {} of property \"{}\" is not a number. Please choose a value in the range [{},{}]".format(\
                            input_text, property, property_hint.property_type_int_min, property_hint.property_type_int_max))
                    return False

        return True


    def on_cancel_button_clicked(self):
        self.was_cancelled = True
        self.close()

    def ok(self):
        return not self.was_cancelled
    
    def cancelled(self):
        return self.was_cancelled
    
    def on_path_button_clicked(self, button):
        directory_path = QFileDialog.getExistingDirectory(self, 'Select Directory', os.getcwd())
        
        if not directory_path:
            return
        
        clicked_property_name = self.__property_path_buttons[button]

        ui_element = self.__property_input_ui_elements[clicked_property_name]
        line_edit = ui_element[0]
        line_edit.setText(directory_path) 



from PyQt5.QtWidgets import QGraphicsView


class NonScrollableGraphicsView(QGraphicsView):
    def __init__(self, scene):
        super().__init__(scene)

    def scrollContentsBy(self, dx, dy):
        print("Scroll contents by ", dx, dy)
        if(dx != 0):
            return
        # Prevent horizontal scrolling
        super().scrollContentsBy(0, dy)



from PyQt5.QtWidgets import QGraphicsPathItem, QGraphicsSceneHoverEvent, QMenu, QAction
from PyQt5.QtGui import QPainterPath, QPen, QColor, QBrush
from PyQt5.QtCore import Qt, pyqtSignal

from ui.module_graphics_item import ModuleGraphicsItem


class PathGraphicsItem(QGraphicsPathItem):

    def __init__(self, path: QPainterPath, parent=None):
        self.hovered = False

        if parent != None:
            super(PathGraphicsItem, self).__init__(parent)
        else:
            super(PathGraphicsItem, self).__init__()

        #TODO: Would it make sense to add createPath and calculatePath methods to this class?

        # Initialize major path attributes
        self.setPath(path)

        # Set colors and thickness
        self.thickness = 2.5
        self.default_color = QColor("#2C6EBE")
        self.default_pen = QPen(self.default_color, self.thickness)
        self.default_brush = QBrush(self.default_color)
        self.hover_color = QColor("#BBDEFB")
        self.hover_pen = QPen(self.hover_color, self.thickness * 1.5)
        self.hover_brush = QBrush(self.hover_color)
        self.select_color = QColor("#37474F")
        self.select_pen = QPen(self.select_color, self.thickness * 1.5)
        self.select_brush = QBrush(self.select_color)

        # Input and output modules
        self.input_module = None
        self.output_module = None

        # Move path in front of other QGraphicsItems
        # Explanation: Hosts and paths are both childs of the scene
        # Thus, you can use z-values to control the order of these items
        self.setZValue(1)

        # Set flags
        self.setFlag(QGraphicsPathItem.ItemIsSelectable, True)
        self.setFlag(QGraphicsPathItem.ItemIsFocusable, True)
        self.setAcceptHoverEvents(True)

    def hoverEnterEvent(self, event):
        self.hovered = True
        self.update()
        super(PathGraphicsItem, self).hoverEnterEvent(event)

    def hoverLeaveEvent(self, event):
        self.hovered = False
        self.update()
        super(PathGraphicsItem, self).hoverLeaveEvent(event)
    
    def keyPressEvent(self, event):
        if (event.key() == Qt.Key_Delete) or (event.key() == Qt.Key_Backspace):
            self.deleteItem()
    
    def contextMenuEvent(self, event):
        self.context_menu = QMenu()
        delete_action = QAction("Delete", self.context_menu)
        delete_action.triggered.connect(self.deleteItem)
        self.context_menu.addAction(delete_action)
        self.context_menu.popup(event.screenPos())

    def deleteItem(self):
        # Remove paths from all lists and dicts
        self.output_module.delete_path(self)
        self.input_module.delete_path(self)

        # Remove path from scene
        scene = self.scene()
        scene.removeItem(self)

        # Delete path
        del self
        
    def paint(self, painter, option, widget):
        # Set color and brush
        if self.isSelected():
            path_pen = self.select_pen
            path_brush = self.select_brush
        elif self.isHovered():
            path_pen = self.hover_pen
            path_brush = self.hover_brush
        else:
            path_pen = self.default_pen
            path_brush = self.default_brush
        painter.setPen(path_pen)
        painter.setBrush(path_brush)

        # Draw path
        painter.drawPath(self.path())

    def get_input_module(self) -> ModuleGraphicsItem:
        return self.input_module
    
    def get_output_module(self) -> ModuleGraphicsItem:
        return self.output_module
    
    def set_input_module(self, module: ModuleGraphicsItem) -> None:
        self.input_module = module

    def set_output_module(self, module: ModuleGraphicsItem) -> None:
        self.output_module = module

    def isHovered(self) -> bool:
        return self.hovered

    def set_hovered(self, hovered: bool) -> None:
        self.hovered = hovered


import sys
import os
current_file_path = os.path.abspath(__file__)

# Get the directory containing the currently executed Python file
current_directory = os.path.dirname(current_file_path)

sys.path.append(current_directory)

from claid_designer.claid_designer import CLAIDDesigner


from PyQt5.QtWidgets import QSpacerItem, QTabWidget, QLabel, QHeaderView, QFrame, QTableWidget, QTableWidgetItem, QWidget, QGraphicsWidget, QMessageBox, QErrorMessage, QInputDialog, QLineEdit, QApplication, QGraphicsScene, QGraphicsView, QGraphicsProxyWidget, QTextEdit, QVBoxLayout, QHBoxLayout, QGraphicsRectItem, QGraphicsLineItem, QDialog, QSizePolicy, QListWidget, QPushButton, QGraphicsOpacityEffect
from PyQt5.QtCore import QPointF, Qt, QLineF, QSize, QRectF, QPropertyAnimation, QPoint, QPropertyAnimation, QSequentialAnimationGroup, QParallelAnimationGroup, pyqtSignal
from PyQt5.QtGui import QResizeEvent, QFontMetrics, QMouseEvent, QFont, QPixmap

import os
import ipaddress
import sys

from ui.animated_button import AnimatedButton


class LandingPage(QWidget):
        
    on_start_clicked = pyqtSignal(str)


    def __init__(self):
        super().__init__()

        self.init_ui()

        self.ip = ""

    def init_ui(self):
        # Set up the main layout
        layout = QVBoxLayout(self)

        # Set the size to 1/4 of the screen width and height
        screen_rect = QApplication.desktop().screenGeometry()
        self.width = screen_rect.width() // 10 * 7
        self.height = screen_rect.height() // 10 * 7

        initial_x = (screen_rect.width() - self.width) // 2
        initial_y = (screen_rect.height() - self.height) // 2

        self.setGeometry(initial_x, initial_y, self.width, self.height)

        # Set the blue background
        self.setStyleSheet("background-color: #2094f3;")

        # Remove menu bar and toolbar
        self.setWindowFlag(Qt.FramelessWindowHint)
        self.setWindowFlag(Qt.WindowCloseButtonHint)

        layout.addStretch(2)
        title = QLabel("From Code to Wrist: Innovating Healthcare with the CLAID Middleware for ML Deployment", self)
        title.setFont(QFont("Arial", 20, QFont.Bold, False))
        title.setAlignment(Qt.AlignCenter | Qt.AlignTop)
        title.setStyleSheet("color: #FFFFFF")
        layout.addWidget(title)
        layout.addStretch(5)

        label = QLabel("CLAID: Closing the Loop on AI & Data Collection", self)
        label.setFont(QFont("Arial", 36, QFont.Bold, False))
        label.setAlignment(Qt.AlignCenter)
        label.setStyleSheet("color: #FFFFFF")
        layout.addWidget(label)
        layout.addStretch(2)

         # Centered a rounded Start button
        self.start_button = AnimatedButton("Begin")
        self.start_button.setStyleSheet("QPushButton { font-family: Arial, sans-serif; font-size: 18px; font-weight: bold; background-color: #F03434; color: white; border: none; border-radius: 10px; }"
                                    "QPushButton:hover { background-color: #FF4E4E; }"
                                    "QPushButton:pressed { background-color: #C0392B; }")
        self.start_button.setMinimumWidth(200)
        self.start_button.setMinimumHeight(50)
        self.start_button.clicked.connect(self.start_button_clicked)
        layout.addWidget(self.start_button, alignment=Qt.AlignCenter)

        # Create a horizontal layout to contain the text, icon, and link horizontally
        info_layout = QHBoxLayout()

        # Informational text label
        info_layout.addStretch()
        info_text = QLabel("For more information, visit our website:")
        info_text.setFont(QFont("Arial", 16))
        info_text.setStyleSheet("color: #FFFFFF")
        info_layout.addWidget(info_text)

        # Icon label
        icon_label = QLabel(self)

        current_file_path = os.path.abspath(__file__)
        # Get the directory containing the currently executed Python file
        current_directory = os.path.dirname(current_file_path)
        globe_icon = QPixmap(os.path.join(current_directory, "icons/globe.png"))
        icon_label.setPixmap(globe_icon.scaledToHeight(20, Qt.SmoothTransformation))
        info_layout.addWidget(icon_label)

        # Link label
        link_label = QLabel(self)
        link_label.setText('<a href="https://claid.ethz.ch" '
                           'style="color: #FFFFFF; text-decoration: underline;">'
                           'claid.ch</a>')
        link_label.setOpenExternalLinks(True)
        link_label.setFont(QFont("Arial", 16))
        # linkLabel.setStyleSheet("margin-top: 20px;")  # Add some space above the link
        link_label.setAlignment(Qt.AlignCenter)
        info_layout.addWidget(link_label)
        info_layout.addStretch()

        # Add the horizontal layout to the main vertical layout
        layout.addLayout(info_layout)
        layout.addStretch(1)



        self.line_edit = QLineEdit(self)
        self.line_edit.setPlaceholderText("Enter text...")
        self.line_edit.setAlignment(Qt.AlignCenter)
        self.line_edit.setMinimumWidth(200)
        self.line_edit.setMinimumHeight(25)
        self.line_edit.returnPressed.connect(self.start_button_clicked)
        self.line_edit.setStyleSheet("QLineEdit { background-color: white; border-radius: 5px; }")
        # self.line_edit.move(self.start_button.pos())

        self.opacity_effect = QGraphicsOpacityEffect(self.line_edit)
        self.opacity_effect.setOpacity(0.0)  # Set initial opacity to 0.0
        self.line_edit.setGraphicsEffect(self.opacity_effect)
        layout.addWidget(self.line_edit, alignment=Qt.AlignCenter)
        layout.addStretch(6)
       

        # Add pictures and speakers side by side at the bottom 
        picture_layout = QHBoxLayout()
        current_path = os.path.abspath(os.path.dirname(__file__))

        # Example pictures (replace these paths with your actual picture paths)
        picture1 = QLabel(self)
        pixmap1 = QPixmap(os.path.join(current_path, "images", "eth_logo.png"))
        picture1.setPixmap(pixmap1.scaledToHeight(100, Qt.SmoothTransformation))
        picture1.setAlignment(Qt.AlignBottom | Qt.AlignLeft)
        picture_layout.addWidget(picture1)

        speakers = QLabel("""<b>Speakers:</b><br>
                          Patrick Langer (planger@ethz.ch)<br>
                          Filipe Barata (fbarata@ethz.ch)<br>
                          Jinjoo Shim (jshim@ethz.ch)<br>
                          Fan Wu (fanwu@ethz.ch)<br>
                          Stephan Altmüller (stephan@klaravision.com)""", self)
        speakers.setFont(QFont("Arial", 16))
        speakers.setStyleSheet("QLabel { color: #FFFFFF; border: 2px solid white; border-radius: 10px; padding: 10px; }")
        speakers.setAlignment(Qt.AlignBottom | Qt.AlignCenter)
        picture_layout.addWidget(speakers)

        picture2 = QLabel(self)
        pixmap2 = QPixmap(os.path.join(current_path, "images", "amld_logo.png"))
        picture2.setPixmap(pixmap2.scaledToHeight(170, Qt.SmoothTransformation))
        picture2.setAlignment(Qt.AlignBottom | Qt.AlignRight)
        picture_layout.addWidget(picture2)
        layout.addLayout(picture_layout)

        # Add the close button ("X") to the top right corner
        close_button = QPushButton("X", self)
        close_button.setGeometry(self.width - 30, 10, 20, 20)
        close_button.clicked.connect(self.close_exit)
        close_button.setStyleSheet("QPushButton { background-color: none; color: white; border: none; }"
            "QPushButton:hover { background-color: #D3D3D3; }"
            "QPushButton:pressed { background-color: #A9A9A9; }"
        )

        self.begin_clicked = False


    def mousePressEvent(self, event: QMouseEvent):
        if event.button() == Qt.LeftButton:
            self.dragging = True
            #TODO: The dragging produces an error if you click on button and move the mouse outside button
            #TODO: Same for the close button
            self.offset = event.pos()

    def mouseMoveEvent(self, event: QMouseEvent):
        if self.dragging:
            self.move(self.pos() + event.pos() - self.offset)

    def mouseReleaseEvent(self, event: QMouseEvent):
        if event.button() == Qt.LeftButton:
            self.dragging = False

    def close_window(self):
        self.close()

    def close_exit(self):
        os._exit(0)

    def start_button_clicked(self):

        if self.begin_clicked:

            ip_address_str = self.line_edit.text()
            valid = False
            exception = None
            try:
                ip, port = ip_address_str.split(":")
                port = int(port)
                addr = ipaddress.ip_address(ip)
                valid = True
            except Exception as e:
                valid = False
                exception = e

          
            if not valid:
                QMessageBox.critical(None, "Invalid IP4 Address", "The value \"{}\" is not a valid IPv4 address:\n{}".format(ip_address_str, str(exception)))
                return

            self.ip = ip_address_str
            # self.on_start_clicked.emit(ip_address_str)
            self.close_window()
            return
        
        self.begin_clicked = True
        # Animate the appearance of the QLineEdit
        self.line_edit.setHidden(False)
        self.line_edit.clear()
        self.animate_line_edit()

    def animate_line_edit(self):
        
        self.setFocus()
        start_button_center = self.start_button.geometry().center()
        start_pos = QPoint((start_button_center.x() - self.line_edit.width() // 2), self.start_button.geometry().top())
        end_pos = QPoint(start_pos.x(), start_pos.y() - self.start_button.geometry().height())
        self.line_edit.setPlaceholderText("Enter Smartwatch IP:Port")
        self.animation = QPropertyAnimation(self.line_edit, b'pos')
        self.animation.setStartValue(start_pos)
        self.animation.setEndValue(end_pos)
        self.animation.setDuration(100)  # Adjust the duration as needed

        self.start_button.setText("Start CLAID")
        # Animate the opacity of the QGraphicsOpacityEffect
        self.opacity_animation = QPropertyAnimation(self.line_edit.graphicsEffect(), b'opacity')
        self.opacity_animation.setStartValue(0.0)
        self.opacity_animation.setEndValue(100)
        self.opacity_animation.setDuration(100)  # Adjust the duration as needed

        # Create a sequential animation group for coordinated animations
        self.animation_group = QParallelAnimationGroup()
        self.animation_group.addAnimation(self.animation)
        self.animation_group.addAnimation(self.opacity_animation)
        self.animation_group.start()

    def get_ip(self):
        return self.ip


from ui.claid_designer_ui import CLAIDDesignerUI
from PyQt5.QtCore import Qt, QPointF, QEvent, QCoreApplication
from PyQt5.QtWidgets import QApplication
import sys
import os

from ui.landing_page import LandingPage

class CLAIDDesigner:

    def __init__(self):
        self.claid = None
        self.__attached_to_claid = False

        self.ip = ""

    def attach(self, claid):

        if self.__attached_to_claid:
            raise RuntimeError("Already attached to claid")
        
        # Those two functions will be called later by CLAID once it has started.
        claid.register_external_function_on_claid_thread_run_once(self.__init_application)
        claid.register_external_function_on_claid_thread_repeat_indefinitely(self.__update)

        self.claid = claid
        
        self.__attached_to_claid = True

    def start(self, host_name, user_id, device_id, module_factory):
        
        self.app = QApplication(sys.argv)

        landing_page = LandingPage()
        # self.landing_page.on_start_clicked.connect(self.on_landing_page_start_clicked)
        landing_page.show()

        while landing_page.isVisible():
            self.__update()

        self.ip = landing_page.get_ip()

        current_file_path = os.path.abspath(__file__)

        # Get the directory containing the currently executed Python file
        current_directory = os.path.dirname(current_file_path)

        self.__replace_in_file(os.path.join(current_directory, "empty_workshop_config.json"), "config.claid", "$(workshop_smartwatch_ip_to_replace)", self.ip)
    
        self.claid.start("config.claid", host_name, user_id, device_id, module_factory)    

        self.available_modules = self.claid.get_available_modules_on_this_host()



    def __init_application(self):
        #TODO: Scale the whole application depending on screen resolution
        self.claid.enable_designer_mode()
        self.mainWindow = CLAIDDesignerUI(self.claid, self.ip)

        # Get screen resolution
        screen = self.app.primaryScreen()
        rect = screen.availableGeometry()
        screenWidth = rect.width()
        screenHeight = rect.height()

        # Window size
        winWidth = int(screenWidth // 1.4)
        winHeight = int(screenHeight // 1.6)

        # Calculate window position
        x = (screenWidth - winWidth) // 2
        y = (screenHeight - winHeight) // 2

        self.mainWindow.setGeometry(x, y, winWidth, winHeight)
        self.mainWindow.setWindowTitle("CLAID Designer")
        self.mainWindow.show()

        


    def __update(self):
        QCoreApplication.processEvents()


    def __on_connected_to_server(self):
        pass

    def __on_disconnected_from_server(self):
        pass

    def __replace_in_file(self, file_path, output_file_path, old_string, new_string):
        try:
            # Open the file for reading
            with open(file_path, 'r') as file:
                file_content = file.read()

            # Replace the specified string
            modified_content = file_content.replace(old_string, new_string)

            # Open the file for writing and overwrite its content
            with open(output_file_path, 'w') as file:
                file.write(modified_content)

            print(f"String '{old_string}' replaced with '{new_string}' in {file_path}")
        except FileNotFoundError:
            print(f"File not found: {file_path}")
        except Exception as e:
            print(f"An error occurred: {str(e)}")


from PyQt5.QtWidgets import QWidget, QVBoxLayout, QPushButton, QGraphicsDropShadowEffect
from PyQt5.QtCore import QPropertyAnimation, QRect

class AnimatedButton(QPushButton):
    def __init__(self, text):
        super().__init__(text)

        # Set dimensions
        self.defaultSize = None
        self.defaultPos = None
        self.growth = None

        # Create shadow
        shadow = QGraphicsDropShadowEffect(self)
        shadow.setBlurRadius(15)
        shadow.setOffset(0, 0)
        self.setGraphicsEffect(shadow)

        # Create animation
        self.animation = QPropertyAnimation(self, b"geometry")
        self.animation.setDuration(30)  # Animation duration in milliseconds

    def enterEvent(self, event):
        super().enterEvent(event)
        # Set the default size and position at the first hover event
        if self.defaultSize == None:
            self.defaultSize = self.size()
        if self.defaultPos == None:
            self.defaultPos = self.pos()
        if self.growth == None:
            self.growth = self.width() * 0.02
        # Animate the button
        self.animation.stop()
        self.animation.setStartValue(self.geometry())
        self.animation.setEndValue(QRect(self.defaultPos.x() - self.growth/2, self.defaultPos.y() - self.growth/2, \
                                        self.width() + self.growth, self.height() + self.growth))
        self.animation.start()

    def leaveEvent(self, event):
        super().leaveEvent(event)
        # Animate the button
        self.animation.stop()
        self.animation.setStartValue(self.geometry())
        self.animation.setEndValue(QRect(self.defaultPos, self.defaultSize))
        self.animation.start()




import sys
from pathlib import Path
from PyQt5.QtWidgets import QApplication, QWidget, QMainWindow, QGraphicsScene, QGraphicsView, QVBoxLayout, QHBoxLayout, QGraphicsPathItem, QPushButton, QFileDialog, QMessageBox, QAction
from PyQt5.QtGui import QPainter, QPainterPath, QColor, QPen, QBrush, QWindow, QKeySequence, QPolygonF
from PyQt5.QtCore import Qt, QPointF, QEvent, QCoreApplication

from ui.module_catalog import ModuleCatalog

import os
sys.path.append(os.getcwd() + "/dispatch/python")

from ui.host_graphics_item import HostGraphicsItem
from ui.module_graphics_item import ModuleGraphicsItem
from ui.channel import Channel
from ui.path_graphics_item import PathGraphicsItem

from claid import CLAID
from claid.module.module_factory import ModuleFactory
from claid.dispatch.proto.claidservice_pb2  import CLAIDConfig, HostConfig, ModuleConfig, ModuleAnnotation, DataPackage, DataFile, ModuleInjectionDescription, Runtime, ConfigUploadPayload
from config.config import Config

from claid.dispatch.proto.claidservice_pb2 import LogMessage, LogMessageSeverityLevel


class CLAIDDesignerUI(QMainWindow):
    def __init__(self, claid, server_ip):
        super(CLAIDDesignerUI, self).__init__()     

        self.claid = claid
        self.server_ip = server_ip

        self.claid.register_on_connected_to_server_callback(self.__on_connected_to_server)
        self.claid.register_on_disconnected_from_server_callback(self.__on_disconnected_from_server)

        # No channel clicked at beginning
        self.firstClickedChannel = None

        # Set flags
        QApplication.instance().installEventFilter(self)

        # Save all hosts and paths in a dictionary
        self.hosts = dict()
        self.host_injected_modules = dict()
        self.paths = dict()

        # Initialize the UI 
        self.initUI()

        # Set the close event
        self.closeEvent = self.exit_application


    def exit_application(self, event):
        os._exit(0)

    def initUI(self):
        self.scene = QGraphicsScene()
        self.view = QGraphicsView(self.scene)
        self.view.setRenderHint(QPainter.Antialiasing, True)

        # Add the view to the layout
        self.layout = QVBoxLayout()
        self.layout.addWidget(self.view)

        # Set a central widget
        self.centralWidget = QWidget()
        self.centralWidget.setLayout(self.layout)
        self.setCentralWidget(self.centralWidget)

        # Create button to start programming of the device
        self.program_device_button = QPushButton("Program devices")

        current_path = os.path.abspath(os.path.dirname(__file__))
        style_sheet_path = os.path.join(current_path, 'style.qss')
        stylesheet = Path(style_sheet_path).read_text()
        self.program_device_button.setStyleSheet(stylesheet)
        self.program_device_button.setFixedWidth(500)
        hbox = QHBoxLayout() 
        hbox.addStretch(1) 
        hbox.addWidget(self.program_device_button)
        hbox.addStretch(1)
        self.layout.addLayout(hbox)
        self.program_device_button.clicked.connect(self.program_device_button_clicked)

        # Create menu bar
        menubar = self.menuBar()
        file_menu = menubar.addMenu("File")

        open_config = QAction("Open Configuration", self)
        open_config.setShortcut(QKeySequence.Open)
        open_config.triggered.connect(self.openConfig)
        file_menu.addAction(open_config)

        save_config = QAction("Save Configuration", self)
        save_config.setShortcut(QKeySequence.Save)
        save_config.triggered.connect(self.saveConfig)
        file_menu.addAction(save_config)

        # Create two hosts, define their properties, and add them to the scene
        host1 = HostGraphicsItem("Smartwatch", 0, 0, 550, 450, self.claid)
        host2 = HostGraphicsItem("Laptop", 600, 0, 550, 450, self.claid)
        

        
        host1.set_type("android")
        host1.set_is_server(True)
        host1.set_server_address(self.server_ip)
        host1.set_min_log_severity_level(LogMessageSeverityLevel.INFO)
        host1.set_supports_module_injections_out_of_host(False)

        host2.set_type("laptop")
        host2.set_connect_to("Smartwatch")
        host2.set_is_server(False)
        host2.set_min_log_severity_level(LogMessageSeverityLevel.WARNING)
        host2.set_supports_module_injections_out_of_host(True)

        host1.onChannelClicked.connect(self.onChannelClicked)
        host1.onModuleItemMoved.connect(self.onModuleItemMoved)
        host1.onInjectableModuleMoveOutsideHost.connect(self.onInjectableModuleMovedOutsideHostSlot)

        host2.onChannelClicked.connect(self.onChannelClicked)
        host2.onModuleItemMoved.connect(self.onModuleItemMoved)
        host2.onInjectableModuleMoveOutsideHost.connect(self.onInjectableModuleMovedOutsideHostSlot)

        self.hosts["Smartwatch"] = host1
        self.hosts["Laptop"] = host2

        #TODO: delete module instances as soon as modulecatalog is available
        # host1.add_module("Module1", "type1", 175, 60, ["input1", "input2", "input3", "input4"], \
        # ["output1", "output2", "output3"])
        # host1.add_module("Module2", "type2", 175, 190, ["input1", "input2"], ["output1", "output2"])
        # host1.add_module("Module3", "type2", 175, 320, ["input1", "input2"], ["output1", "output2"])
        # host2.add_module("Module4", "type1", 175, 60, ["input1", "input2"], ["output1", "output2"])
        # host2.add_module("Module5", "type2", 175, 190, ["input1", "input2"], ["output1", "output2"])
        # host2.add_module("Module6", "type3", 175, 320, ["input1", "input2"], ["output1", "output2"])

        self.scene.addItem(host1)
        self.scene.addItem(host2)

        self.__on_disconnected_from_server()



    def onChannelClicked(self, channel):
        if self.firstClickedChannel == None:
            self.firstClickedChannel = channel
        else:
            self.createPath(self.firstClickedChannel, channel)
            self.firstClickedChannel = None

    def eventFilter(self, source, event):
        # Ensures that self.firstClickedChannel is set to None when clicking on the scene
        if isinstance(source, QWindow):
            return super().eventFilter(source, event)
        if event.type() == QEvent.MouseButtonPress:
            if event.button() == Qt.LeftButton or event.button() == Qt.RightButton:
                pos = self.view.mapToScene(event.pos())
                # Find the item at the clicked position
                items = self.scene.items(pos)
                if items:
                    channel_found = None
                    for item in items:
                        if item != None and isinstance(item, Channel):
                            channel_found = item
                            break                     
                    if channel_found == None:
                        self.firstClickedChannel = None
        return super().eventFilter(source, event)
    
    def makeUniqueChannelTuple(self, channel1: Channel, channel2: Channel):
        # Sort the channels as following: channel1 is output channel, channel2 is input channel
        # Alternatively: Sort the channels based on memory address, hence making
        # This function work regardless of the order the channels were passed.
        if channel1 in channel1.parentItem().input_channel_map.values():
            channel1,channel2 = channel2,channel1
        
        return (channel1, channel2)

    # Draw path between two channels
    def createPath(self, channel1: Channel, channel2: Channel):
        # Check if channels are children of the same module
        if channel1.parentItem() == channel2.parentItem():
            return
        
        # Ensure that channel1 is an output channel and channel2 an input channel
        if channel1 in channel1.parentItem().output_channel_map.values() and channel2 in channel2.parentItem().output_channel_map.values():
            return
        elif channel1 in channel1.parentItem().input_channel_map.values() and channel2 in channel2.parentItem().input_channel_map.values():
            return
        elif channel1 in channel1.parentItem().input_channel_map.values():
            temp_channel = channel1
            channel1 = channel2
            channel2 = temp_channel

        print("Checking type")
        if not channel1.has_same_data_type_as_channel(channel2):
            QMessageBox.critical(None, "Invalid channel types", "Cannot connect channel \"{}\" to channel \"{}\" due to incompatible channel types.\n\n"\
                                 "Channel \"{}\" has data type \"{}\", but Channel \"{}\" has data type \"{}\".".format(channel1.get_name(), channel2.get_name(),\
                                                                                                                        channel1.get_name(), channel1.get_data_type_name(),
                                                                                                                        channel2.get_name(), channel2.get_data_type_name()))
            return
    
        # Check if this path is already existing
        if self.pathExists(channel1, channel2):
            return

        # Calculate the path
        path = self.calculatePath(channel1, channel2)         
        
        # Create a QGraphicsPathItem
        pathItem = PathGraphicsItem(path)

        # Stack the path before hosts
        for host in self.hosts.values():
            pathItem.stackBefore(host)

        # Add the path to the scene
        self.scene.addItem(pathItem)
        
        # Add the path to the path dictionary in current class && ModuleGraphicsItem
        # && Add the modules to the PathGraphicsItem
        self.paths[self.makeUniqueChannelTuple(channel1, channel2)] = pathItem
        channel1.parentItem().add_path(pathItem)
        channel2.parentItem().add_path(pathItem)
        pathItem.set_output_module(channel1.parentItem())
        pathItem.set_input_module(channel2.parentItem())

        # Connect a signal that is emitted on deletion of the path
        channel1.parentItem().pathDeleted.connect(self.delete_path)

    def calculatePath(self, channel1: Channel, channel2: Channel):
        #TODO: NEED TO ADJUST PATH IN EDGE CASES TO INCLUDE FULL ARROWS
        # Ensure that channel1 is an output channel and channel2 an input channel
        if channel1 in channel1.parentItem().input_channel_map.values():
            temp_channel = channel1
            channel1 = channel2
            channel2 = temp_channel
       
        # Create a QPainterPath
        path = QPainterPath()

        # Calculate start and end points
        ch1_x = channel1.boundingRect().right()
        ch2_x = channel2.boundingRect().left()
        ch1_y = channel1.boundingRect().top() + (channel1.boundingRect().height() / 2)
        ch2_y = channel2.boundingRect().top() + (channel2.boundingRect().height() / 2) 
        start = channel1.mapToScene(ch1_x, ch1_y)
        end = channel2.mapToScene(ch2_x, ch2_y)

        # Minimum offset between modules and path
        offset = 10

        # Arrow head length
        arrow_length = 8

        # Case 1: Output channel module is to the left of the input channel module
        if (channel1.scenePos().x() + channel1.parentItem().circle_diameter) <= channel2.scenePos().x():
            # Calculate midpoints
            midX = (start.x() + end.x()) / 2
            midY1 = start.y()
            midY2 = end.y()
            midpoint1 = QPointF(midX, midY1)
            midpoint2 = QPointF(midX, midY2)

            # Construct the path
            path.moveTo(start)
            path.lineTo(midpoint1)
            path.moveTo(midpoint1)
            path.lineTo(midpoint2)
            path.moveTo(midpoint2)
            path.lineTo(end)

        # Case 2: Output channel module is to the right of the input channel module
        else:
            # Case 2_1: Output channel module is positioned above input channel module
            if channel1.parentItem().scenePos().y() <= channel2.parentItem().scenePos().y():
                # Case 2_1_1 Only slightly above
                if (channel1.parentItem().scenePos().y() + channel1.parentItem().boundingRect().bottom()) > (channel2.parentItem().scenePos().y() - (2 * offset)):
                    # Case 2_1_1_1: Output channel module is positioned to the left of input channel module
                    if channel1.parentItem().scenePos().x() < channel2.parentItem().scenePos().x():
                        # Calculate midX1
                        midX1 = start.x() + (channel2.parentItem().scenePos().x() - channel1.parentItem().scenePos().x()) - (channel1.parentItem().circle_diameter / 2) + offset
                    # Case 2_1_1_2: Output channel module is positioned to the right of input channel module
                    else:
                        # Calculate midX1
                        midX1 = start.x() - (channel1.parentItem().circle_diameter / 2) + offset
                    # Calculate all other midpoints
                    midY1 = start.y()
                    midX2 = midX1
                    midY2 = channel2.parentItem().scenePos().y() + channel2.parentItem().boundingRect().bottom() + offset
                    midX3 = channel2.parentItem().scenePos().x() - offset - arrow_length
                    midY3 = midY2
                    midX4 = midX3
                    midY4 = end.y()
                # Case 2_1_2: Significantly above
                else:
                    midX1 = start.x() - (channel1.parentItem().circle_diameter / 2) + offset
                    midY1 = start.y()
                    midX2 = midX1
                    midY2 = ((channel1.parentItem().scenePos().y() + channel1.parentItem().boundingRect().bottom())  + channel2.parentItem().scenePos().y()) / 2
                    midX3 = channel2.parentItem().scenePos().x() - offset - arrow_length
                    midY3 = midY2
                    midX4 = midX3
                    midY4 = end.y()
            # Case 2_2: Output channel module is positioned below input channel module
            else:
                # Case 2_2_1: Only slightly below
                if (channel1.parentItem().scenePos().y() - (2 * offset)) < (channel2.parentItem().scenePos().y() + channel2.parentItem().boundingRect().bottom()):
                    # Case 2_2_1_1: Output channel module is positioned to the left of input channel module
                    if channel1.parentItem().scenePos().x() < channel2.parentItem().scenePos().x():
                        # Calculate midX1
                        midX1 = start.x() + (channel2.parentItem().scenePos().x() - channel1.parentItem().scenePos().x()) - (channel1.parentItem().circle_diameter / 2) + offset
                    # Case 2_2_1_2: Output channel module is positioned to the right of input channel module
                    else:
                        # Calculate midX1
                        midX1 = start.x() - (channel1.parentItem().circle_diameter / 2) + offset
                    # Calculate all other midpoints
                    midY1 = start.y()
                    midX2 = midX1
                    midY2 = channel2.parentItem().scenePos().y() - offset
                    midX3 = channel2.parentItem().scenePos().x() - offset - arrow_length
                    midY3 = midY2
                    midX4 = midX3
                    midY4 = end.y()
                # Case 2_2_2: Significantly below
                else:
                    midX1 = start.x() - (channel1.parentItem().circle_diameter / 2) + offset
                    midY1 = start.y()
                    midX2 = midX1
                    midY2 = (channel1.parentItem().scenePos().y() + (channel2.parentItem().scenePos().y() + channel2.parentItem().boundingRect().bottom())) / 2
                    midX3 = channel2.parentItem().scenePos().x() - offset - arrow_length
                    midY3 = midY2
                    midX4 = midX3
                    midY4 = end.y()

            # Define midpoints
            midpoint1 = QPointF(midX1, midY1)
            midpoint2 = QPointF(midX2, midY2)
            midpoint3 = QPointF(midX3, midY3)
            midpoint4 = QPointF(midX4, midY4)


            # Construct the path
            path.moveTo(start)
            path.lineTo(midpoint1)
            path.moveTo(midpoint1)
            path.lineTo(midpoint2)
            path.moveTo(midpoint2)
            path.lineTo(midpoint3)
            path.moveTo(midpoint3)
            path.lineTo(midpoint4)
            path.moveTo(midpoint4)
            path.lineTo(end)   

        # Add arrow head
        path.moveTo(end)
        path.addPolygon(QPolygonF([end, QPointF((end.x() - arrow_length), (end.y() - (arrow_length / 2))), \
                                   QPointF((end.x() - arrow_length), (end.y() + (arrow_length / 2))), end]))
        path.closeSubpath()

        return path

    # Check if a path already exists
    def pathExists(self, channel1: Channel, channel2: Channel):
        
        path_key = self.makeUniqueChannelTuple(channel1, channel2)

        return path_key in self.paths

    #! Understand how this works again -> can I delete the arguments?
    # Redraw the path if a module is moved
    def onModuleItemMoved(self, module, change):
        # Iterate over all paths
        for entry in self.paths:

            channel1, channel2 = entry
            path_item = self.paths[(channel1, channel2)]

            path = self.calculatePath(channel1, channel2)
            path_item.setPath(path)

    def openConfig(self):
        #* Open file dialog
        downloads_path = os.path.expanduser(os.getcwd())
        while True:
            file_path, _ = QFileDialog.getOpenFileName(self, "Open File", downloads_path, "JSON Files (*.json);; All Files (*)")
            
            # Return if file dialog was cancelled
            if not file_path:
                return
            
            # Check if the selected file is a JSON file
            if not file_path.endswith(".json"):
                QMessageBox.warning(self, "Invalid File", "Please select a JSON file.")
                continue
            
            # Check if the JSON file contains compatible content
            try:
                config = Config()
                config.load_config_from_file(file_path)
            except Exception as e:
                QMessageBox.warning(self, "Error", "An error occurred while loading the file. \
                                    Please check if the JSON file contains a compatible configuration. " + repr(e))
                continue

            # If a valid file was selected, break the loop
            break
        
        reply = QMessageBox.question(self, "Save Configuration", "Unsaved changes will be lost. Do you want to save before continuing?", \
                                     QMessageBox.Yes | QMessageBox.No, QMessageBox.Yes)
        if reply == QMessageBox.Yes:
            self.saveConfig()

        config = Config()
        config.load_config_from_file(file_path)

        # Check if host is online && modules are available
        for host in config.get_config_data().hosts:
            host_name = host.hostname
            available_modules = self.claid.get_available_modules_for_host(host_name)
            # Check if host is online
            if available_modules == None:
                QMessageBox.critical(None, f"\"{host_name}\" not connected.", 
                                     (f"Cannot open this configuration. \"{host_name}\" is not connected. "
                                     f"Please make sure that \"{host_name}\" is connected before opening this configuration."))
                return
            for module in host.modules:
                if module.type not in available_modules.keys():
                    QMessageBox.warning(self, "Invalid Module Type", "Cannot load config with module type " + module.type + ". This module tpye is unknown to CLAID.")
                    return

        #TODO: (Better call a delete configuration function that should also delete all paths)
        #* Delete old configuration
        # Delete hosts (will automatically delete mdodules and channels, too)
        for host in self.hosts.values():
            self.scene.removeItem(host)
        self.hosts.clear()

        # Delete paths
        for path in self.paths.values():
            self.scene.removeItem(path)
        self.paths.clear()

        #* Create hosts and modules
        # Positions of first host and first module in this host
        x_host = 0
        y_host = 0
        x_module = 175
        y_module = 60

        for host in config.get_config_data().hosts:
            # Get host data
            host_name = host.hostname
            host_type = host.type
            host_is_server = host.is_server
            host_server_address = host.host_server_address
            host_connect_to = host.connect_to
            host_min_log_severity_level = host.min_log_severity_level
            available_modules = self.claid.get_available_modules_for_host(host_name)
            
            #TODO: Delete the is_connected check if I solved this check differently
            # if not self.hosts[host_name].is_connected():
            #     QMessageBox.critical(None, f"{host_name} not connected.", 
            #                          (f"Cannot open this configuration. {host_name} is not connected. "
            #                          f"Please make sure that {host_name} is connected before opening this configuration."))
            #     return
            
            # Calculate dimensions of host
            host_width = 550
            num_modules = len(host.modules)
            if num_modules <= 3:
                host_height = 450
            else:
                host_height = 450 + (num_modules - 3) * 130
            
            # Create a HostGraphicsItem
            host_graphics_item = HostGraphicsItem(host_name, x_host, y_host, host_width, host_height, self.claid)
            host_graphics_item.onChannelClicked.connect(self.onChannelClicked)
            host_graphics_item.onModuleItemMoved.connect(self.onModuleItemMoved)
            
            # Set all host properties
            host_graphics_item.set_type(host_type)
            if host_is_server:
                host_graphics_item.set_is_server(host_is_server)
            if host_server_address != None:
                host_graphics_item.set_server_address(host_server_address)
            if host_connect_to != None:
                host_graphics_item.set_connect_to(host_connect_to)
            host_graphics_item.set_min_log_severity_level(host_min_log_severity_level)
            if(host_name == "Laptop"):
                host_graphics_item.set_supports_module_injections_out_of_host(True)
     
            for module in host.modules:
                #! Falls die connection genau vor oder während dem Auslesen der "available_modules" unterbrochen, bräuchte ich nochmal eine Abfrage, 
                #! ob available_modules != None && module.type in available_modules.keys()??? Oder kann ich das weglassen?
                # Get module data
                # if module.type not in available_modules.keys():
                #     QMessageBox.warning(self, "Invalid Module Type", "Cannot load config with module type " + module.type + ". This module tpye is unknown to CLAID.")
                #     return
                module_id = module.id
                module_type = module.type
                module_properties = module.properties
                module_annotation = available_modules[module_type]
                module_input_channels = self.claid.get_input_channels_of_module(module_annotation)
                module_output_channels = self.claid.get_output_channels_of_module(module_annotation)

                module_input_channel_types = self.claid.get_input_channel_types_of_module(module_annotation)
                module_output_channel_types = self.claid.get_output_channel_types_of_module(module_annotation)

                module_property_names = module_annotation.properties
                module_property_descriptions = module_annotation.property_descriptions
                module_hints = module_annotation.property_hints

                # Create a ModuleGraphicsItem 
                module_graphics_item = host_graphics_item.add_module(module_id, module_type, x_module, y_module, module_input_channels, module_output_channels, \
                                              module_input_channel_types, module_output_channel_types, module_property_names, module_property_descriptions, module_hints)
                
                # Set properties
                module_graphics_item.set_properties(module_properties)
                
                # Set y position for next module
                y_module += 130
            
            # Store host and its modules in the dictionary && add the hosts to the scene
            self.hosts[host.hostname] = host_graphics_item
            self.scene.addItem(host_graphics_item)

            x_host += 650
            y_module = 60

        #* Re-Initialize MainWindow and UI
        # No channel clicked at beginning
        self.firstClickedChannel = None
                
        #TODO: There needs to be a dialog in case for uncertain cases (connection vs no connection)
        #* Create paths
        # Step 1: Iterate over connected output channels to find each connection id
        channel_descriptions = config.get_channel_descriptions()

        for connection in channel_descriptions:

            publishers = channel_descriptions[connection].publisher_modules
            subscribers = channel_descriptions[connection].subscriber_modules

            for pub in publishers:
                pub_host, pub_module, pub_channel = pub

                for sub in subscribers:
                    sub_host, sub_module, sub_channel = sub

                    channel1 = None
                    channel2 = None

                    #TODO: Add dict (host, module, channel) <-> ChannelItem
                    for target_out_module in self.hosts[pub_host].get_modules():
                        if target_out_module.id == pub_module:
                            channel1 = target_out_module.output_channel_map[pub_channel]
                            break
                    for target_in_module in self.hosts[sub_host].get_modules():
                        if target_in_module.id == sub_module:
                            channel2 = target_in_module.input_channel_map[sub_channel]
                            break

                    if channel1 != None and channel2 != None:
                        # Step 2: Create paths between the two found channels
                        self.createPath(channel1, channel2)

    def saveConfig(self):
        #* Open file dialog
        downloads_path = os.path.expanduser(os.getcwd())
        file_path, _ = QFileDialog.getSaveFileName(self, "Save File", downloads_path, "JSON Files (*.json);; All Files (*)")
        
        # Return if file dialog was cancelled
        if not file_path:
            return
        
        # Add .json extension to the file path if it doesn't have one
        if not file_path.endswith(".json"):
            file_path += ".json"

        claid_config = self.generateCLAIDConfigFromUI()

        config = Config()   
        Config.save_config_to_file(claid_config, file_path)        

    def generateCLAIDConfigFromUI(self):
        #* Build a "connections" dict that contains tuple (hostname, module_id, channel_name) and its corresponding connection name
        # Create empty dictionary for connections
        connections = dict()

        # Initial connection number, used for naming the connections
        connect_num = 0

        # Iterate over all paths
        for (channel1, channel2) in self.paths:
            h1 = channel1.parentItem().parentItem().get_name()
            h2 = channel2.parentItem().parentItem().get_name()
            m1 = channel1.parentItem().get_id()
            m2 = channel2.parentItem().get_id()
            c1 = channel1.get_name()
            c2 = channel2.get_name()

            # Case1: channel1 (output channel) already in connections
            if (h1, m1, c1) in connections:
                if (h2, m2, c2) not in connections:
                    connections[(h2, m2, c2)] = connections[(h1, m1, c1)]
                elif connections[(h2, m2, c2)] != connections[(h1, m1, c1)]:
                    for (h, m, c) in connections:
                        if connections[(h, m, c)] == connections[(h2, m2, c2)]:
                            connections[(h, m, c)] = connections[(h1, m1, c1)]
                else:
                    continue
            
            # Case 2: channel1 (output channel) not in connections && channel2 (input channel) already in connections
            elif (h2, m2, c2) in connections:
                connections[(h1, m1, c1)] = connections[(h2, m2, c2)]
            
            # Case 3: channel1 (output channel) and channel2 (input channel) not in connections
            else:
                connection_name = "connection" + str(connect_num)
                connections[(h1, m1, c1)] = connection_name
                connections[(h2, m2, c2)] = connection_name
                connect_num += 1

        # for (host, module, channel) in connections:
        #     print("Host: " + host + ", Module: " + module + ", Channel: " + channel + ", Connection: " + connections[(host, module, channel)])

        #* Save hosts, modules and connections into config
        claid_config = CLAIDConfig()
        
        for host in self.hosts.values():
            host_config = HostConfig()
            
            # Gather all host properties:
            host_name = host.get_name()
            host_type = host.get_type()
            host_is_server = host.get_is_server()
            host_server_address = host.get_server_address()
            host_connect_to = host.get_connect_to()
            host_min_log_severity_level = host.get_min_log_severity_level()
        
            # Save all host properties into host_config
            host_config.hostname = host_name
            host_config.type = host_type
            if host_is_server:
                host_config.is_server = host_is_server
            if host_server_address != None:
                host_config.host_server_address = host_server_address
            if host_connect_to != None:
                host_config.connect_to = host_connect_to
            host_config.min_log_severity_level = host_min_log_severity_level

            for module in host.modules:
                module_config = ModuleConfig()

                # Gather all module properties:
                module_id = module.get_id()
                module_type = module.get_type()
                module_properties = module.get_properties()
                module_output_channels = dict()
                module_input_channels = dict()

                for (channel1, channel2) in self.paths:
                    # Check if channel1 (output channel) is child of module
                    if channel1.parentItem() == module:
                        pub_host = host_name
                        pub_module = module_id
                        pub_channel = channel1.get_name()
                        connection_name = connections[(pub_host, pub_module, pub_channel)]

                        # Store connection name. No need to check if name is already stored as the name is unique
                        module_output_channels[pub_channel] = connection_name

                    # Check if channel2 (input channel) is child of module
                    elif channel2.parentItem() == module:
                        sub_host = host_name
                        sub_module = module_id
                        sub_channel = channel2.get_name()
                        connection_name = connections[(sub_host, sub_module, sub_channel)]

                        # Store connection name. No need to check if name is already stored as the name is unique
                        module_input_channels[sub_channel] = connection_name

                # Save all module properties into module_config
                module_config.id = module_id
                module_config.type = module_type
                #TODO: Check if the if condition is really necessary
                if module_properties:
                    module_config.properties.update(module_properties)
                if module_output_channels:
                    module_config.output_channels.update(module_output_channels)
                if module_input_channels:
                    module_config.input_channels.update(module_input_channels)

                # Append module_config to host_config
                host_config.modules.append(module_config)

            # Append host_config to claid_config
            claid_config.hosts.append(host_config)
         
        return claid_config

    def __on_connected_to_server(self):
        
        self.hosts["Smartwatch"].set_connected()

    def __on_disconnected_from_server(self):

        self.hosts["Smartwatch"].set_disconnected()

    def program_device_button_clicked(self):
        print("program device button clicked")

        if not self.hosts["Smartwatch"].is_connected():
            QMessageBox.critical(None, "Smartwatch not connected", "Cannot program devices. Smartwatch is not connected.")
            return
        
        reply = QMessageBox.question(self, "Program devices", "Devices are ready. Upload configuration now?", QMessageBox.Yes | QMessageBox.No, QMessageBox.Yes)
        if reply == QMessageBox.Yes:

            if not self.hosts["Smartwatch"].is_connected():
                QMessageBox.critical(None, "Smartwatch not connected", "Cannot program devices. Smartwatch is not connected.")
                return
            
            config = self.generateCLAIDConfigFromUI()

            self.claid.upload_config_to_host(self.claid.get_current_host_id(), config)

            payload = self.get_injection_payload_for_host(self.hosts["Smartwatch"])
            self.claid.upload_config_to_host_with_payload("Smartwatch", config, payload)

    def get_injection_payload_for_host(self, host):

        # TODO: Generalize this. This currently only works for Smartwatch<->Host, and assumes
        # the injected module is available locally.
        payload = ConfigUploadPayload()
        for module in host.get_modules():
            if module.was_injected():
    
                module_type = module.get_type()
                module_python_file_path = self.claid.get_module_factory().get_path_to_python_file_of_module(module_type)

                module_file_name = os.path.basename(module_python_file_path)

                payload_files = []

                module_data_file = self.load_file_into_data_file(module_python_file_path)
                module_data_file.relative_path = module_file_name
                payload_files.append(module_data_file)

                injection_description = ModuleInjectionDescription()
                injection_description.module_name = module.get_type()
                injection_description.module_file = module_file_name
                injection_description.runtime = Runtime.RUNTIME_PYTHON

                print("Path of injected module: ", self.claid.get_module_factory().get_path_to_python_file_of_module(module_type))
                
                module_folder_path = os.path.dirname(module_python_file_path)

                
                print("dependencies: \n")
                for dependency in module.get_injection_file_dependencies():
                    print(str(dependency) + "\n")
                    data_file = self.load_file_into_data_file(os.path.join(module_folder_path, dependency))
                    data_file.relative_path = data_file.relative_path.replace(module_folder_path, "")
                    payload_files.append(data_file)


                payload.payload_files.MergeFrom(payload_files)
                payload.modules_to_inject.MergeFrom([injection_description])
            
        print(payload)
        return payload

    def load_file_into_data_file(self, file_path):
        data_file = DataFile()

        # Set relative_path to the provided file path
        data_file.relative_path = file_path

        # Read file data and assign it to file_data field
        with open(file_path, 'rb') as file:
            data_file.file_data = file.read()

        return data_file


    def delete_path(self, del_path):
        for (channel1, channel2) in self.paths:
            if self.paths[(channel1, channel2)] == del_path:
                del self.paths[(channel1, channel2)]
                break

    def get_hosts(self):
        return self.hosts
    
    def onInjectableModuleMovedOutsideHostSlot(self, host, module):
        print("designer moved outside")

        target_host = self.hosts["Smartwatch"]
        if module.is_inside_host(target_host):

            module_dependencies = module.get_injection_file_dependencies()

            dependency_text = ""
            if len(module_dependencies) > 0:
                dependency_text = "The Module has the following dependencies, which will be injected to host \"{}\" as well:\n".format(target_host.get_hostname())

                for dependency in module_dependencies:
                    dependency_text += "\"" + dependency + "\"\n"

            reply = QMessageBox.question(None, "Inject Module?", "Do you want to inject Module \"{}\" from host \"{}\" into host \"{}\"?\n{}".format(\
                module.get_id(), host.get_hostname(), target_host.get_hostname(), dependency_text), QMessageBox.Yes | QMessageBox.No, QMessageBox.Yes)
            
            if reply == QMessageBox.Yes:
                
                module.setParentItem(target_host)
                module.set_is_injectable(False)
                module.set_was_injected(True)

                target_host.inject_module(module)
                host.remove_module(module)
                QMessageBox.information(None, "Module injected", "Module \"{}\" was injected to host \"{}\". It will be uploaded the next time you click on \"Program Devices\"."\
                                        .format(module.get_id(), target_host.get_hostname()))
                return
           
        center_point = host.mapToScene(host.rect().center())

        # Calculate the new position for rect1 to be in the middle of rect2
        new_x = center_point.x() - module.rect().width() / 2
        new_y = center_point.y() - module.rect().height() / 2

        # Set the new position for rect1
        module.setPos(new_x, new_y)


from claid import CLAID
from claid.module.module_factory import ModuleFactory

from claid_designer import CLAIDDesigner

claid = CLAID()
designer = CLAIDDesigner()

designer.attach(claid)
designer.start("Laptop", "laptop_user", "device", ModuleFactory())
# claid.start("empty_workshop_config.json", "Laptop", "laptop_user", "device", ModuleFactory())



from pathlib import Path
from PyQt5.QtWidgets import QMessageBox, QWidget, QGraphicsItem, QGraphicsObject, QVBoxLayout, QGraphicsEllipseItem, QGraphicsTextItem, QPushButton, QGraphicsProxyWidget, QGraphicsPixmapItem
from PyQt5.QtGui import QPen, QColor, QPainterPath, QIcon, QPixmap
from PyQt5.QtCore import Qt, QRectF, pyqtSignal, QPointF
from ui.module_catalog import ModuleCatalog
from claid.dispatch.proto.claidservice_pb2 import LogMessage, LogMessageSeverityLevel

from ui.module_graphics_item import ModuleGraphicsItem
import os
import sip
class HostGraphicsItem(QGraphicsObject):

    onHostItemMoved = pyqtSignal(QGraphicsObject, QPointF)
    onHostItemResized = pyqtSignal(int, int)
    onAddModuleButtonClicked = pyqtSignal(QGraphicsObject)

    onModuleItemMoved = pyqtSignal(QGraphicsObject, QPointF)
    onChannelClicked = pyqtSignal(QGraphicsObject)

    onInjectableModuleMoveOutsideHost = pyqtSignal(QGraphicsObject, QGraphicsObject) 

    def __init__(self, hostname: str, x: int, y: int, width: int, height: int, claid, parent=None):
        if parent != None:
            super(HostGraphicsItem, self).__init__(parent)
        else:
            super(HostGraphicsItem, self).__init__()

        # Initialize major host attributes
        self.hostname = hostname
        self.type = ""
        self.is_server = False
        self.server_address = ""
        self.connect_to = ""
        self.min_log_severity_level = LogMessageSeverityLevel.INFO
        self.modules = list()

        # If true, the ModuleCatalog will support setting the is_injectable flag on Modules,
        # if they are marked as injectable in their ModuleAnnotation
        self.module_injections_out_of_host_supported = False
        self.claid = claid

        # Set dimensions of modules
        self.module_width = 150
        self.module_height = 100

        # Set border color
        self.color = QColor("#A5C2E8")        

        # Create a QRectF object that represents the host
        self.rectf = QRectF(x, y, width, height)
        self.draw = True
        self.to_draw = self.rectf

        # Create resizeHandle
        self.resizeHandle = QGraphicsEllipseItem(x + self.rect().width() - 10,  y + self.rect().height() - 10, 10, 10, self)
        self.resizeHandle.setBrush(Qt.darkGray)
        self.resizing = False

        self.mouseMove = False

        # Create add_module_button
        current_path = os.path.abspath(os.path.dirname(__file__))
        self.add_module_button = QPushButton(QIcon(os.path.join(current_path, "icons/add_white.svg")), "Add module")
        current_path = os.path.abspath(os.path.dirname(__file__))
        style_sheet_path = os.path.join(current_path, 'style.qss')
        stylesheet = Path(style_sheet_path).read_text()
        self.add_module_button.setStyleSheet(stylesheet)
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.add_module_button)
        widget = QWidget()
        widget.setLayout(layout)
        proxy = QGraphicsProxyWidget(self)
        proxy.setWidget(widget)
        proxy.setPos((x + 10), (y + 30))
        self.add_module_button.clicked.connect(self.on_add_module_button_clicked)

        # Set flags
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)
        self.setAcceptHoverEvents(True)

        self.connected = True

        current_path = os.path.abspath(os.path.dirname(__file__))

        self.icon_connected = QPixmap(os.path.join(current_path, "icons/wifi_connected.png")).scaled(30, 30, Qt.KeepAspectRatio, Qt.SmoothTransformation)
        self.icon_disconnected = QPixmap(os.path.join(current_path, "icons/wifi_disconnected.png")).scaled(30, 30, Qt.KeepAspectRatio, Qt.SmoothTransformation)
        self.icon_item = QGraphicsPixmapItem(self.icon_connected, self)
        #TODO: VERY IMPORTANT: SET THE POSITION DEPENDING ON TEXT HEIGHT IN PAINTER METHOD
        self.icon_item.setPos(self.rectf.right() - self.icon_connected.width() - 5, self.rectf.top() + 25)


    def on_add_module_button_clicked(self):
        if not self.connected:
            QMessageBox.critical(None, "Error", "Cannot currently add Modules to host \"{}\". Host is not connected.".format(self.hostname))
            return 
        
        self.module_catalog = ModuleCatalog(self, self.claid, self.hostname, self.module_injections_out_of_host_supported)
        self.module_catalog.show()
        self.module_catalog.on_catalog_closed.connect(self.on_catalog_closed)

    def on_catalog_closed(self):
        print("Catalog closed")
        sip.delete(self.module_catalog)

    def add_module(self, module_id: str, module_type: str, x: int, y: int, input_channels: list, output_channels: list, 
                   input_channel_types: dict, output_channel_types: dict, property_names: list, property_descriptions: list, property_hints: list):
        
        module_graphics_item = ModuleGraphicsItem(module_id, module_type, x, y, self.module_width, self.module_height, \
                                                  input_channels, output_channels, input_channel_types, output_channel_types, \
                                                    property_names, property_descriptions, property_hints, self)
        self.modules.append(module_graphics_item)
    
        for channel in module_graphics_item.input_channel_map.values():
            channel.onChannelClicked.connect(self.onChannelClickedSlot)
        for channel in module_graphics_item.output_channel_map.values():
            channel.onChannelClicked.connect(self.onChannelClickedSlot)
        
        module_graphics_item.onModuleItemMoved.connect(self.onModuleItemMovedSlot)
        module_graphics_item.onInjectableModuleMoveOutsideHost.connect(self.onInjectableModuleMoveOutsideHostSlot)


        return module_graphics_item
    
    def inject_module(self, module_graphics_item):
        self.modules.append(module_graphics_item)
    
        for channel in module_graphics_item.input_channel_map.values():
            channel.onChannelClicked.disconnect()
            channel.onChannelClicked.connect(self.onChannelClickedSlot)
        for channel in module_graphics_item.output_channel_map.values():
            channel.onChannelClicked.disconnect()
            channel.onChannelClicked.connect(self.onChannelClickedSlot)
        
        module_graphics_item.onModuleItemMoved.disconnect()
        module_graphics_item.onModuleItemMoved.connect(self.onModuleItemMovedSlot)

        module_graphics_item.onInjectableModuleMoveOutsideHost.disconnect()
        module_graphics_item.onInjectableModuleMoveOutsideHost.connect(self.onInjectableModuleMoveOutsideHostSlot)
    
    def remove_module(self, module):

        for mod in self.modules:
            if mod == module:
                self.modules.remove(module)
                return



    def onChannelClickedSlot(self, channel):
        self.onChannelClicked.emit(channel)

    def onModuleItemMovedSlot(self, graphics_item, change):
        self.onModuleItemMoved.emit(graphics_item, change)

    def boundingRect(self):
        return self.rectf
    
    def rect(self):
        return self.rectf
    
    def setRect(self, x, y, new_width, new_height):
        self.draw = False
        self.to_draw = QRectF(x, y, new_width, new_height)
        self.update()

    def paint(self, painter, option, widget):

        if self.draw:
            self.rectf = self.to_draw
            self.repaint(painter)
        else:
            self.draw = True
            self.update()

    def mousePressEvent(self, event):
        # Clear selection of other GraphicsItems
        if not self.isSelected():
            self.scene().clearSelection()

        if self.resizeHandle.isUnderMouse():
            self.resizing = True
        elif event.button() == Qt.LeftButton:
            self.mousePressPos = event.pos()
            self.origPos = self.pos()
            self.mouseMove = False
        super(HostGraphicsItem, self).mousePressEvent(event)

    def mouseMoveEvent(self, event):
        if event.buttons() == Qt.LeftButton and self.resizing:
            self.resizeItem(event.pos())
        elif event.buttons() == Qt.LeftButton:
            self.mouseMove = True
            super(HostGraphicsItem, self).mouseMoveEvent(event)
        else:
            super(HostGraphicsItem, self).mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        if self.resizing:
            self.resizing = False
        if event.button() == Qt.LeftButton:
            if not self.mouseMove:
                # Handle click event here if needed
                pass
        super(HostGraphicsItem, self).mouseReleaseEvent(event)

    def hoverMoveEvent(self, event):
        if self.resizeHandle.isUnderMouse():
            self.setCursor(Qt.SizeFDiagCursor)
        else:
            self.setCursor(Qt.ArrowCursor)
        super(HostGraphicsItem, self).hoverMoveEvent(event)

    #TODO: Ensure that host always contains all modules + some offset
    def resizeItem(self, pos):
        new_width = max(pos.x() - self.rect().x(), 10)
        new_height = max(pos.y() - self.rect().y(), 10)
        self.setRect(self.rect().x(), self.rect().y(), new_width, new_height)
        self.resizeHandle.setRect(self.rect().x() + self.rect().width() - 10, self.rect().y() + self.rect().height() - 10, 10, 10)
        self.onHostItemResized.emit(new_width, new_height)

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemPositionChange:
            new_pos = value
            # TODO: Implement restrictMovement to ensure that host objects cannot overlap
            # changed_pos = self.restrictMovement(new_pos)
            absolute_change = new_pos - self.pos()
            self.onHostItemMoved.emit(self, absolute_change)
            return super(HostGraphicsItem, self).itemChange(change, new_pos)
        else:
            return super(HostGraphicsItem, self).itemChange(change, value)

    def repaint(self, painter):
        # Set border thickness
        border_thickness = 3

        # Draw rectangle
        corner_radius = 10
        rect = self.rect()
        painter.setPen(QPen(self.color, border_thickness))
        painter.setBrush(Qt.white)
        painter.drawRoundedRect(rect, corner_radius, corner_radius)

        # Draw smaller rectangle at the top for the text
        text_dimensions = painter.boundingRect(QRectF(), Qt.AlignLeft, self.hostname)   
        text_rect_height = 1.35 * text_dimensions.height()
        text_rect = QRectF(rect.x(), rect.y(), rect.width(), text_rect_height)
        path = QPainterPath()
        path.moveTo(text_rect.left(), text_rect.top() + corner_radius)
        path.arcTo(text_rect.left(), text_rect.top(), 2 * corner_radius, 2 * corner_radius, 180, -90)
        path.lineTo(text_rect.right() - corner_radius, text_rect.top())
        path.arcTo(text_rect.right() - 2 * corner_radius, text_rect.top(), 2 * corner_radius, 2 * corner_radius, 90, -90)
        path.lineTo(text_rect.right(), text_rect.bottom())
        path.lineTo(text_rect.left(), text_rect.bottom())
        path.closeSubpath()
        painter.setBrush(self.color)
        painter.drawPath(path)

        # Set color for text
        painter.setPen(Qt.white)

        # Draw text of host name
        x = rect.x()
        y = rect.y()
        width = rect.width()
        x_position = x + (width / 2) - (text_dimensions.width() / 2)
        y_position = y + text_dimensions.height()
        painter.drawText(int(x_position), int(y_position), self.hostname)        

    def get_hostname(self):
        return self.hostname
    

    def get_color(self, ) -> QColor:
        color = self.color
        return color

    def set_color(self, color: QColor) -> None:
        self.color = color
        # Use this color for errors or warnings: 
        self.update()

    def set_name(self, hostname: str):
        self.hostname = hostname
        self.update()

    def get_name(self) -> str:
        return self.hostname
    
    def set_type(self, type: str):
        self.type = type

    def get_type(self) -> str:
        return self.type

    def set_is_server(self, is_server: bool):
        self.is_server = is_server

    def get_is_server(self) -> bool:
        return self.is_server

    def set_server_address(self, server_address: str):
        self.server_address = server_address

    def get_server_address(self) -> str:
        return self.server_address
    
    def set_connect_to(self, connect_to: str):
        self.connect_to = connect_to
    
    def get_connect_to(self) -> str:
        return self.connect_to

    def get_modules(self):
        return self.modules
    
    def does_module_with_name_exist(self, module_name: str):

        for module in self.modules:
            if module.get_id() == module_name:
                return True

        return False
    
    def delete_module(self, module_id: str) -> None:
        for module in self.modules:
            if module.get_id() == module_id:
                self.modules.remove(module)
    
    def set_connected(self):
        self.set_color(QColor("#A5C2E8"))
        self.connected = True
        self.icon_item.setPixmap(self.icon_connected)

    def set_disconnected(self):
        self.set_color(QColor("#A50000"))
        self.connected = False
        self.icon_item.setPixmap(self.icon_disconnected)

    def is_connected(self):
        return self.connected


    def get_min_log_severity_level(self):
        return self.min_log_severity_level
    
    def set_min_log_severity_level(self, severity_level):
        self.min_log_severity_level = severity_level

    def onInjectableModuleMoveOutsideHostSlot(self, module):
        print("Host: module moved outside")
        self.onInjectableModuleMoveOutsideHost.emit(self, module)

    def set_supports_module_injections_out_of_host(self, supports_injections):
        self.module_injections_out_of_host_supported = supports_injections
    
    def supports_module_injections_out_of_host(self):
        return self.module_injections_out_of_host_supported


from PyQt5.QtWidgets import QMessageBox, QGraphicsItem, QGraphicsObject, QGraphicsTextItem, QMenu, QAction, QApplication
from PyQt5.QtGui import QFontMetrics, QColor, QPen, QBrush, QPainterPath, QFont
from PyQt5.QtCore import Qt, QRectF, pyqtSignal, QPointF
from ui.property_input_dialog import PropertyInputDialog

from ui.channel import Channel
import sip


class ModuleGraphicsItem(QGraphicsObject):
    
    onModuleItemMoved = pyqtSignal(QGraphicsObject, QPointF)
    pathDeleted = pyqtSignal(QGraphicsItem)
    onInjectableModuleMoveOutsideHost = pyqtSignal(QGraphicsObject) 

    def __init__(self, id: str, type: str, x: int, y: int, width: int, height: int, input_channels: list, output_channels: list, \
                 input_channel_types: dict, output_channel_types: dict, property_names: list, property_descriptions: list, property_hints: list, parent=None, movable=True):
        #! Could I also move this downwoards?
        self.movable = movable
        self.hovered = False

        if parent != None:
            super(ModuleGraphicsItem, self).__init__(parent)
        else:
            super(ModuleGraphicsItem, self).__init__()
        
        # Initialize major module attributes
        self.id = id
        self.type = type
        self.input_channel_map = dict()
        self.output_channel_map = dict()
        self.input_channel_text_map = dict()
        self.output_channel_text_map = dict()
        self.properties = dict()

        self.property_names = property_names
        self.property_descriptions = property_descriptions
        self.property_hints = property_hints

        self.injectable = False
        self.injected = False

        # Set dimensions
        self.width = width
        self.height = height

        # Set colors and thickness:
        self.thickness = 2
        self.corner_radius = 10
        self.default_color = QColor("#2C6EBE")
        self.default_pen = QPen(self.default_color, self.thickness)
        self.default_brush = QBrush(self.default_color)
        self.hover_color = QColor("#BBDEFB")
        self.hover_pen = QPen(self.hover_color, self.thickness * 1.5)
        self.hover_brush = QBrush(self.hover_color)
        self.select_color = QColor("#37474F")
        self.select_pen = QPen(self.select_color, self.thickness * 1.5)
        self.select_brush = QBrush(self.select_color)

        # #TODO: Just for demonstration purpose. Delete later
        # if self.id == "Module3":
        #     self.set_default_color(QColor("#D32F2F"))

        # Set position && connect signals
        if parent != None:
            parent_rect = self.parentItem().rect()
            self.setPos(parent_rect.left() + x, parent_rect.top() + y)
            
            parent.onHostItemResized.connect(self.onHostItemResized)
            parent.onHostItemMoved.connect(self.onHostItemMoved)
       
        # Set input and output channels
        self.set_channels(input_channels, output_channels, input_channel_types, output_channel_types)

        # List of connected paths
        self.paths = list()

        # Set flags
        self.mouseMove = False
        if movable:
            self.setFlag(QGraphicsItem.ItemIsMovable, True)
            self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)
            self.setAcceptHoverEvents(True)
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)
        self.setFlag(QGraphicsItem.ItemIsFocusable, True)

        self.item_was_moved = False

        self.injection_file_dependencies = []

    def boundingRect(self):
        return self.rectf
    
    def rect(self):
        return self.rectf
    
    def setRect(self, x, y, new_width, new_height):
        self.rectf = QRectF(x, y, new_width, new_height)
        self.update()

    def hoverEnterEvent(self, event):
        path = QPainterPath()
        path.addRoundedRect(self.rect(), self.corner_radius, self.corner_radius)
        if path.contains(event.pos()):
            self.hovered = True
            self.update()
        super(ModuleGraphicsItem, self).hoverEnterEvent(event)

    def hoverMoveEvent(self, event):
        path = QPainterPath()
        path.addRoundedRect(self.rect(), self.corner_radius, self.corner_radius)
        if path.contains(event.pos()):
            self.hovered = True
            self.update()
        super(ModuleGraphicsItem, self).hoverMoveEvent(event)

    def hoverLeaveEvent(self, event):
        path = QPainterPath()
        path.addRoundedRect(self.rect(), self.corner_radius, self.corner_radius)
        if not path.contains(event.pos()):
            self.hovered = False
            self.update()
        super(ModuleGraphicsItem, self).hoverLeaveEvent(event)
    
    def mousePressEvent(self, event):
        # Give focus to the item when clicked
        self.setFocus()
        self.item_was_moved = False
        # Clear selection of other GraphicsItems
        if not self.isSelected():
            self.scene().clearSelection()

        if event.button() == Qt.LeftButton:
            self.mousePressPos = event.pos()
            self.origPos = self.pos()
            self.mouseMove = False
        super(ModuleGraphicsItem, self).mousePressEvent(event)

    def mouseMoveEvent(self, event):
        if event.buttons() == Qt.LeftButton:
            self.mouseMove = True
        super(ModuleGraphicsItem, self).mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        if event.button() == Qt.LeftButton:

            self.check_if_module_was_moved_outside_host()

            if not self.mouseMove:
                # Handle click event here if needed
                pass
        super(ModuleGraphicsItem, self).mouseReleaseEvent(event)

    def check_if_module_was_moved_outside_host(self):
        print("mouse release event, ", self.item_was_moved)

        if(not self.is_inside_host(self.parentItem())):
            self.onInjectableModuleMoveOutsideHost.emit(self)
  
    def is_inside_host(self, host):
        bounding_rect1 = self.mapToScene(self.rect()).boundingRect()
        bounding_rect2 = host.mapToScene(host.rect()).boundingRect()
        return bounding_rect2.contains(bounding_rect1)


    def itemChange(self, change, value):
        if not self.movable:
            return super(ModuleGraphicsItem, self).itemChange(change, value)

        if change == QGraphicsItem.ItemPositionChange:
            new_pos = value
            changed_pos = self.restrictMovement(new_pos)
            absolute_change = changed_pos - self.pos()
            self.onModuleItemMoved.emit(self, absolute_change)
            self.item_was_moved = True
            return super(ModuleGraphicsItem, self).itemChange(change, changed_pos)
        else:
            return super(ModuleGraphicsItem, self).itemChange(change, value)
    
    def keyPressEvent(self, event):
        if (event.key() == Qt.Key_Delete) or (event.key() == Qt.Key_Backspace):
            self.deleteItem()
    
    def contextMenuEvent(self, event):
        self.context_menu = QMenu()
        delete_action = QAction("Delete", self.context_menu)
        delete_action.triggered.connect(self.deleteItem)
        properties_action = QAction("Show Properties", self.context_menu)
        properties_action.triggered.connect(self.showProperties)
        self.context_menu.addAction(delete_action)
        self.context_menu.addAction(properties_action)
        self.context_menu.popup(event.screenPos())

    def deleteItem(self):
        # Ask for confirmation
        reply = QMessageBox.question(None, "Delete module", "Do you really want to delete this module and its connections to other modules?",
                                     QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
        if reply == QMessageBox.No:
            return

        # Remove modules from all lists
        self.parentItem().delete_module(self.id)

        # Delete paths
        while self.paths:
            path = self.paths[0]
            path.deleteItem()

        # Remove module from scene
        scene = self.scene()
        scene.removeItem(self)

        # Delete module
        del self
    
    def showProperties(self):
        empty_properties = [42]

        print("Show properties ", self.property_names, len(empty_properties), len(self.property_names))
        defined_properties = self.properties

        if len(self.property_names) == 0:
            QMessageBox.information(None, "No properties", "This Module has no properties.")
            return

  
        self.property_dialog = PropertyInputDialog(self.property_names, self.property_descriptions, self.property_hints, defined_properties=self.properties)
        self.property_dialog.onOkButtonClicked.connect(self.onPropertyDialogOkClicked)
        self.property_dialog.open()

    def onPropertyDialogOkClicked(self):
       
        defined_properties = self.property_dialog.get_defined_properties()
        empty_properties = self.property_dialog.get_empty_properties()

        if len(empty_properties) != 0:

            empty_list_msg = ""
            for property in empty_properties:
                empty_list_msg += "\t\"" + str(property) +  "\"\n"
            QMessageBox.critical(None, "Undefined property", "You did not specify a value for the following properties:\n{}".format(empty_list_msg) + 
                                    "Please make sure that you define each property.")

            self.property_dialog = PropertyInputDialog(self.property_names, self.property_descriptions, self.property_hints, defined_properties=self.properties)
            self.property_dialog.onOkButtonClicked.connect(self.onPropertyDialogOkClicked)
            self.property_dialog.open()
            return

        self.properties = defined_properties
        


    
    def restrictMovement(self, new_pos):
        if not self.movable:
            return new_pos
        
        if self.injectable:
            return new_pos

        parent_rect = self.parentItem().rect()
        item_rect = self.boundingRect()

        # Restrict movement horizontally
        if new_pos.x() < parent_rect.left():
            new_pos.setX(parent_rect.left())
        elif new_pos.x() + item_rect.width() > parent_rect.right():
            new_pos.setX(parent_rect.right() - item_rect.width())

        # Restrict movement vertically
        if new_pos.y() < parent_rect.top():
            new_pos.setY(parent_rect.top())
        elif new_pos.y() + item_rect.height() > parent_rect.bottom():
            new_pos.setY(parent_rect.bottom() - item_rect.height())

        #TODO: Prevent that modules can collide
        # self_scene_rect = self.mapRectToScene(item_rect)
        # self_scene_rect.moveTo(new_pos)
        # for module in self.parentItem().childItems():
        #     if isinstance(module, ModuleGraphicsItem):
        #         if module != self:
        #             module_scene_rect = module.mapRectToScene(module.boundingRect().marginsAdded(QMarginsF(1, 1, 1, 1)))
        #             if self_scene_rect.intersects(module_scene_rect):
        #                 new_pos = self.pos()

        return new_pos

    def onHostItemResized(self, new_width, new_height):
        self.setPos(self.restrictMovement(self.pos()))
        self.update()

    def onHostItemMoved(self, host, change):
        absolute_change = change
        self.onModuleItemMoved.emit(self, absolute_change)

    def paint(self, painter, option, widget):
        self.repaint(painter)

    def repaint(self, painter):
        if self.isSelected():
            module_pen = self.select_pen
            module_brush = self.select_brush
            for child in self.childItems():
                if isinstance(child, Channel):
                    child.set_default_color(self.select_color)
                    child.update()
        elif self.isHovered():
            module_pen = self.hover_pen
            module_brush = self.hover_brush
            for child in self.childItems():
                if isinstance(child, Channel):
                    child.set_default_color(self.hover_color)
                    child.update()
        else:
            module_pen = self.default_pen
            module_brush = self.default_brush
            for child in self.childItems():
                if isinstance(child, Channel):
                    child.set_default_color(self.default_color)
                    # Attention: child.update() is not called explicitly here 
                    # bcs it seems to be called anyways and would cause a crash
                    # when module catalog is opened after adding a module 
        
        # Draw rectangle
        rect = self.rect()
        painter.setPen(module_pen)
        painter.setBrush(Qt.white)
        painter.drawRoundedRect(rect, self.corner_radius, self.corner_radius)

        #TODO: Need to check ModuleAnnotations if module is Java/C++/... and then draw the text accordingly
        # Draw smaller rectangle at the top for the module_type text
        x = rect.x()
        y = rect.y()
        width = rect.width()
        height = rect.height()
        text_dimensions_type = painter.boundingRect(QRectF(), Qt.AlignLeft, self.type)#+ " | Java")
        text_rect_height = 1.35 * text_dimensions_type.height()
        text_rect = QRectF(x, y, width, text_rect_height)
        path = QPainterPath()
        path.moveTo(text_rect.left(), text_rect.top() + self.corner_radius)
        path.arcTo(text_rect.left(), text_rect.top(), 2 * self.corner_radius, 2 * self.corner_radius, 180, -90)
        path.lineTo(text_rect.right() - self.corner_radius, text_rect.top())
        path.arcTo(text_rect.right() - 2 * self.corner_radius, text_rect.top(), 2 * self.corner_radius, 2 * self.corner_radius, 90, -90)
        path.lineTo(text_rect.right(), text_rect.bottom())
        path.lineTo(text_rect.left(), text_rect.bottom())
        path.closeSubpath()
        painter.setBrush(module_brush)
        painter.drawPath(path)

        # Set color for module_type text
        painter.setPen(Qt.white)

        # Draw module_type text
        text_dimensions_type = painter.boundingRect(QRectF(), Qt.AlignLeft, self.type) #+ " | Java")
        x_position = x + (width / 2) - (text_dimensions_type.width() / 2)
        y_position = y + text_dimensions_type.height()
        painter.drawText(int(x_position), int(y_position), self.type) #+ " | Java")        
        
        # Set color for module_id text
        painter.setPen(Qt.black)

        # Draw module_id text
        text_dimensions_id = painter.boundingRect(QRectF(), Qt.AlignLeft, self.id)
        x_position = x + (width / 2) - (text_dimensions_id.width() / 2)
        y_position = y + ((height + text_rect_height) / 2)
        painter.drawText(int(x_position), int(y_position), self.id)

    def add_path(self, path):
        self.paths.append(path)

    def delete_path(self, path):
        self.paths.remove(path)
        self.pathDeleted.emit(path)

    def get_color(self) -> QColor:
        color = self.default_color
        return color

    def set_default_color(self, color: QColor) -> None:
        self.default_color = color
        self.default_pen = QPen(self.default_color, self.thickness)
        self.default_brush = QBrush(self.default_color)
        # Use this color for errors or warnings: #D32F2F
        self.update()

    def set_id(self, name: str):
        self.id = name
        self.update()

    def get_id(self) -> str:
        return self.id
    
    def set_type(self, type: str):
        self.type = type
        #TODO: When this is also drawn in the module (via (re)paint function), I need to call update() 
        # self.update()

    def get_type(self) -> str:
        return self.type

    def set_channels(self, input_channels, output_channels, input_channel_types, output_channel_types) -> None:

        # Create a rectangle that represents the module
        width = self.width
        height = self.height
        self.rectf = QRectF(0, 0, width, height)
        print("mod dim", self.width, self.height)
        for channel in self.input_channel_map.values():
            print("delete")
            sip.delete(channel)
        self.input_channel_map.clear()

        for channel in self.output_channel_map.values():
            sip.delete(channel)
        self.output_channel_map.clear()

        for text in self.input_channel_text_map.values():
            sip.delete(text)
        self.input_channel_text_map.clear()

        for text in self.output_channel_text_map.values():
            sip.delete(text)
        self.output_channel_text_map.clear()

        

        # Calculate position of input and output channels
        self.circle_diameter = min(width, height) * 0.125
        circle_diameter = self.circle_diameter
        input_x_pos = 0 
        output_x_pos = width
        self.num_input_channels = len(input_channels)
        self.num_output_channels = len(output_channels)
        input_y_distance = height / (self.num_input_channels + 1)
        input_y_pos = input_y_distance - (circle_diameter / 2)
        output_y_distance = height / (self.num_output_channels + 1)
        output_y_pos = output_y_distance - (circle_diameter / 2)
        text_x_offset = 10
        text_y_offset = 1
        font_size = 10
        adjustment = 4  # Somehow to the text always starts a bit too far to the right (and to the top)

        print("dbg 1")
        for input_channel in input_channels:

            if not input_channel in input_channel_types:
                raise ValueError("Unable to find data type of input channel {}".format(input_channel))
            
            data_type = input_channel_types[input_channel]
            print("dbg 2")

            self.input_channel_map[input_channel] = Channel(input_channel, data_type, input_x_pos, input_y_pos, circle_diameter, self)
            self.input_channel_text_map[input_channel] = QGraphicsTextItem(input_channel, self)
            self.input_channel_text_map[input_channel].setDefaultTextColor(Qt.black)
            text_font = self.input_channel_text_map[input_channel].font()
            text_font.setPointSize(font_size)
            self.input_channel_text_map[input_channel].setFont(text_font)
            text_dim = QFontMetrics(text_font)
            text_width = text_dim.width(self.input_channel_text_map[input_channel].toPlainText())
            text_height = text_dim.height()
            text_x_pos = -text_width - text_x_offset - adjustment
            text_y_pos = input_y_pos - text_height - text_y_offset
            self.input_channel_text_map[input_channel].setPos(text_x_pos, text_y_pos)
            input_y_pos += input_y_distance
        for output_channel in output_channels:
            if not output_channel in output_channel_types:
                raise ValueError("Unable to find data type of input channel {}".format(output_channel))
            
            data_type = output_channel_types[output_channel]

            self.output_channel_map[output_channel] = Channel(output_channel, data_type, output_x_pos, output_y_pos, circle_diameter, self)
            self.output_channel_text_map[output_channel] = QGraphicsTextItem(output_channel, self)
            self.output_channel_text_map[output_channel].setDefaultTextColor(Qt.black)
            text_font = self.output_channel_text_map[output_channel].font()
            text_font.setPointSize(font_size)
            self.output_channel_text_map[output_channel].setFont(text_font)
            text_dim = QFontMetrics(self.output_channel_text_map[output_channel].font())
            text_width = text_dim.width(self.output_channel_text_map[output_channel].toPlainText())
            text_height = text_dim.height()
            text_x_pos = width + text_x_offset - adjustment
            text_y_pos = output_y_pos - text_height - text_y_offset
            self.output_channel_text_map[output_channel].setPos(text_x_pos, text_y_pos)
            output_y_pos += output_y_distance
        
        self.update()

    def get_input_channels(self) -> dict:
        return self.input_channel_map
    
    def get_output_channels(self) -> dict:
        return self.output_channel_map

    def get_properties(self) -> dict:
        return self.properties
    
    def set_properties(self, properties) -> None:
        self.properties = properties
    
    # def get_property_names(self) -> list:
    #     return self.property_names
    
    # def get_property_descriptions(self) -> list:
    #     return self.property_descriptions
    
    # def get_property_hints(self) -> list:
    #     return self.property_hints
    
    def isHovered(self) -> bool:
        return self.hovered
    
    def set_is_injectable(self, injectable):
        self.injectable = injectable

    def is_injectable(self):
        return self.injectable

    def set_was_injected(self, injected):
        self.injected = injected

    def was_injected(self):
        return self.injected

    def set_injection_file_dependencies(self, dependencies: list):
        self.injection_file_dependencies = dependencies

    def get_injection_file_dependencies(self):
        return self.injection_file_dependencies 


    def set_hovered(self, hovered: bool) -> None:
        self.hovered = hovered
        self.update()
    


from setuptools import setup, find_packages

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "claid_designer" / "README.md").read_text()

setup(
    name='claid_designer',
    version='0.0.3',    
    description='Designer for the CLAID framework.',
    url='https://claid.ch',
    author='Patrick Langer (C4DHI)',
    author_email='planger@ethz.ch',
    license='Apache 2',
    packages=find_packages(),
    include_package_data=True,
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=['claid',      
                      'matplotlib==3.5.3',      
                      'pyqt5==5.14.2; platform_system=="Linux"',
                      'pyqt5==5.15.10; platform_system=="Darwin"',
                      'opencv-python==4.8.0.76' 
                      ],

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: Apache Software License',  
        'Operating System :: Android',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',        
        'Operating System :: Unix',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    python_requires='>=3.7',


)



from claid.dispatch.proto.claidservice_pb2 import CLAIDConfig, ModuleAnnotation, DataPackage
import json
from google.protobuf.json_format import Parse, ParseDict, MessageToJson

class CLAID:

    def __init__(self):
        pass

    def __make_input_channel(self, channel_name: str, module_name: str):
        package = DataPackage()
        package.channel = channel_name
        package.target_module = module_name

        return package

    def __make_output_channel(self, channel_name: str, module_name: str):
        package = DataPackage()
        package.channel = channel_name
        package.source_module = module_name

        return package


    def get_available_modules(self):
        modules = dict()

        module_description = ModuleAnnotation()
        module_description.module_description = "This is a Module that allows to record audio data from the Microphone"
        module_description.properties.append("RecordingMode")
        module_description.properties.append("BitRate")
        module_description.properties.append("SamplingRate")

        module_description.property_descriptions.append("Mode for recording data, one of: [Continuous, Chunks]")
        module_description.property_descriptions.append("Bitrate of the audio sample, one of [8, 16, 32]")
        module_description.property_descriptions.append("Sampling Rate, e.g. 44100")
        module_description.channel_definition.append(self.__make_output_channel("AudioDataChannel", "AudioRecorder"))
        module_description.channel_description.append("Channel where the recorded AudioData will be posted to.")
        
        modules["AudioRecorder"] = module_description

        module_description = ModuleAnnotation()
        module_description.module_description = "This is a Module that allows to record data from any channel and stores incoming data on the filesystem"
        module_description.properties.append("fileFormat")
        module_description.properties.append("fileName")
        module_description.properties.append("storagePath")

        module_description.property_descriptions.append("Format/codec for the file. One of [JSON, BINARY, BatchJSON, BatchBBINARY]")
        module_description.property_descriptions.append("Naming convention for the files to store. Can use strftime specifiers, e.g.: my_file\%H_\%M")
        module_description.property_descriptions.append("Storage path for the files, e.g. /sdcard/MyFiles")
        module_description.channel_definition.append(self.__make_input_channel("DataChannel", "DataSaver"))
        module_description.channel_description.append("Channel with incoming data.")
        
        modules["DataSaver"] = module_description

        image_processor_description = ModuleAnnotation()
        image_processor_description.module_description = "Processes images by applying filters and transformations."
        image_processor_description.properties.append("filterType")
        image_processor_description.properties.append("resizeFactor")
        image_processor_description.properties.append("outputFormat")

        image_processor_description.property_descriptions.append("Type of filter to apply, e.g., [Blur, Sharpen, Grayscale].")
        image_processor_description.property_descriptions.append("Factor by which to resize the image, e.g., 0.5 for half the size.")
        image_processor_description.property_descriptions.append("Output format for processed images, e.g., [JPEG, PNG].")
        image_processor_description.channel_definition.append(self.__make_input_channel("InputImages", "ImageProcessor"))
        image_processor_description.channel_description.append("Input channel for images to be processed.")
        image_processor_description.channel_definition.append(self.__make_output_channel("ProcessedImages", "ImageProcessor"))
        image_processor_description.channel_description.append("Output channel for processed images.")

        modules["ImageProcessor"] = image_processor_description

        # Add DataAnalyzer Module
        data_analyzer_description = ModuleAnnotation()
        data_analyzer_description.module_description = "Analyzes incoming data and extracts key insights or statistics."
        data_analyzer_description.properties.append("analysisType")
        data_analyzer_description.properties.append("threshold")
        data_analyzer_description.properties.append("outputMetrics")

        data_analyzer_description.property_descriptions.append("Type of analysis to perform, e.g., [Descriptive, Predictive, Diagnostic].")
        data_analyzer_description.property_descriptions.append("Threshold value for specific analyses.")
        data_analyzer_description.property_descriptions.append("List of specific metrics to output.")
        data_analyzer_description.channel_definition.append(self.__make_input_channel("InputData", "DataAnalyzer"))
        data_analyzer_description.channel_description.append("Input channel for data to be analyzed.")
        data_analyzer_description.channel_definition.append(self.__make_output_channel("AnalyzedResults", "DataAnalyzer"))
        data_analyzer_description.channel_description.append("Output channel for analyzed results.")

        modules["DataAnalyzer"] = data_analyzer_description

        # Add ParallelProcessor Module
        parallel_processor_description = ModuleAnnotation()
        parallel_processor_description.module_description = "Processes data in parallel with multiple input and output channels."
        parallel_processor_description.channel_definition.append(self.__make_input_channel("InputData1", "ParallelProcessor"))
        parallel_processor_description.channel_description.append("First input channel for parallel processing.")
        parallel_processor_description.channel_definition.append(self.__make_input_channel("InputData2", "ParallelProcessor"))
        parallel_processor_description.channel_description.append("Second input channel for parallel processing.")
        parallel_processor_description.channel_definition.append(self.__make_output_channel("OutputData1", "ParallelProcessor"))
        parallel_processor_description.channel_description.append("First output channel for parallel processing.")
        parallel_processor_description.channel_definition.append(self.__make_output_channel("OutputData2", "ParallelProcessor"))
        parallel_processor_description.channel_description.append("Second output channel for parallel processing.")

        modules["ParallelProcessor"] = parallel_processor_description

        # Add DataTransformer Module
        data_transformer_description = ModuleAnnotation()
        data_transformer_description.module_description = "Transforms data with multiple input and output channels."
        data_transformer_description.channel_definition.append(self.__make_input_channel("InputData", "DataTransformer"))
        data_transformer_description.channel_description.append("Input channel for data transformation.")
        data_transformer_description.channel_definition.append(self.__make_output_channel("OutputData1", "DataTransformer"))
        data_transformer_description.channel_description.append("First output channel for data transformation.")
        data_transformer_description.channel_definition.append(self.__make_output_channel("OutputData2", "DataTransformer"))
        data_transformer_description.channel_description.append("Second output channel for data transformation.")

        modules["DataTransformer"] = data_transformer_description

        accelerometer_collector_description = ModuleAnnotation()
        accelerometer_collector_description.module_description = "Collects accelerometer data from a device sensor."
        accelerometer_collector_description.properties.append("samplingRate")
        accelerometer_collector_description.properties.append("collectionDuration")

        accelerometer_collector_description.property_descriptions.append("Sampling rate for accelerometer data, e.g., 50 Hz.")
        accelerometer_collector_description.property_descriptions.append("Duration for data collection in seconds.")
        accelerometer_collector_description.channel_definition.append(self.__make_output_channel("AccelerometerData", "AccelerometerCollector"))
        accelerometer_collector_description.channel_description.append("Output channel for collected accelerometer data.")

        modules["AccelerometerCollector"] = accelerometer_collector_description

        return modules
    
    def get_input_channels_of_module(self, module_annotation: ModuleAnnotation):
        channels = list()
        for channel in module_annotation.channel_definition:
            if channel.target_module != "":
                channels.append(channel.channel)

        return channels
    
    

    def get_output_channels_of_module(self, module_annotation: ModuleAnnotation):
        channels = list()
        for channel in module_annotation.channel_definition:
            if channel.source_module != "":
                channels.append(channel.channel)

        return channels

    def save_config_to_file(self, config, json_file_path):
        # Use the .to_dict() method to convert the protobuf message to a dictionary
        json_data = MessageToJson(config, including_default_value_fields=True, preserving_proto_field_name=True)
        print(json_data)
        with open(json_file_path, 'w') as f:
            f.write(json_data)

# message ModuleAnnotation
# {
# string module_description  = 1;
# repeated string properties = 2;
# repeated string property_descriptions = 3;
# repeated DataPackage channel_definition = 4;
# repeated string channel_description = 5;
# }


# from PyQt5.QtWidgets import QGraphicsScene


# # Note: Diese Klasse ist momentan irrelevant. Sie könnte ggf. für customized Elemente später relevant sein
# class DiagramScene(QGraphicsScene):

#     def __init__(self, parent=None):
#         super(DiagramScene, self).__init__(parent)
